/*
 * Author:	K.Sarnath, 2001.
 * License:	GPL.
 *
 * Program demonstrating AVL Trees.
 * Reference: Art of Comp Programming - Donald Knuth.
 * Only Insertion Done. Deletion not yet done : as on : 4th June 2001.
 *
 * I think I can make the algorithm generic and use it wherever possible.
 * Till now, I have not used it for the kernel data structures.
 *                                              As on 18/August/2001
 * Deletion Done: 2/Dec/2001.
 * File last updated on: 2/Dec/2001
 */

#include <asm/unistd.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <fcntl.h>

#define DEBUG_AVL
#define LINUX_APP

#ifdef LINUX_APP
#define printk	printf
#endif

#define ASSERT(x)	\
		if ( !(x) ) {\
			printk(__FUNCTION__"\n Failed condition"#x"\n");\
		}

#define called(x...)	printk("\n"__FUNCTION__":"x);

#define NR_NUM                  9
#define BALANCED                0
#define RIGHT_SKEWED            1
#define RIGHT_UNBALANCED        2
#define LEFT_SKEWED             -1
#define LEFT_UNBALANCED         -2

/*
 * Return values of "balance"
 */
#define HEIGHT_REDUCED 		1
#define HEIGHT_MAINTAINED 	0
#define ILLEGAL			-1


struct AVLnode {
        long data;
        char balance;
        struct AVLnode *left;
        struct AVLnode *right;
        struct AVLnode *parent;
};

#define p(x) x->parent
#define b(x) x->balance

struct AVLnode *root = (struct AVLnode *) NULL;
void AVLinsert ( long, struct AVLnode ** );
struct AVLnode * find( long val, struct AVLnode *root );

#if defined(DEBUG_AVL) &&  defined(LINUX_APP)
void
sync()
{
	asm ("\nmovl %0, %%eax"
	     "\nint $0x80"
	      :
	      : "i" (__NR_sync) 
	      : "ax" );
}
#endif

int main()
{
        int j,i = 0, num;
	struct AVLnode *node;

        i = 0;
        /* Construct the Tree */
       	AVLinsert( 5, &root );
       	AVLinsert( 2, &root );
       	AVLinsert( 6, &root );
       	AVLinsert( 7, &root );
       	AVLinsert( 1, &root );
       	AVLinsert( 4, &root );
       	AVLinsert( 3, &root );

	node = find(7, root); 
	AVLdelete(node, &root);

	traverse_pre_order( root );
	return 0;
}


void
AVLinsert ( long val , struct AVLnode **root )
{
        struct AVLnode *node = (struct AVLnode *) NULL;

	ASSERT(root != NULL);

        node = (struct AVLnode *)malloc(sizeof(struct AVLnode));
        if ( node == NULL ) {
                perror("malloc failed: Reason: ");
                exit (1);
        }

        node->data = val;
        node->balance = BALANCED;
        node->left = NULL;
        node->right = NULL;
        node->parent = NULL;

        if ( *root == NULL )
                *root = node;
        else
                insert_and_balance(node, root);
}

int
insert_and_balance (struct AVLnode *node, struct AVLnode **root)
{
        struct AVLnode *node_in_qn = (struct AVLnode *)NULL;
	struct AVLnode **ptr = (struct AVLnode **) NULL;
        struct AVLnode *parent = (struct AVLnode *)NULL, *place;
        struct AVLnode *X;
        char dir;

insert:
        if ( root == NULL || *root == NULL )
                return -1;

        place = node_in_qn = *root;
/*
 * First, insert him at the right place. Just ordinary binary insert.
 * At the same time, try to identify the position in the binary tree which
 * might need to be balanced. That is defined as :
 * BalanceReqPosn = The Nearest position to the inserted node, who is
 * skewed ( i.e. Bfactor = 1 or -1 ).
 *                                                      -- Sarnath.K
 */
        while ( place ) {
                if ( node->data < place->data ) {
                        ptr = &(place->left);
                        parent = place;
                        place = place->left;
                } else {
                        ptr = &(place->right);
                        parent = place;
                        place = place->right;
                }
                if ( place && ( b(place) != 0 ) )
                        node_in_qn = place;
        }
        *ptr = node;
        node->parent = parent;

/*  
 *  Now, we found out the nearest guy who was skewed.
 *  This data is present in "node_in_qn". Now the PATH from the
 *  "node_in_qn" to the NODE_ADDED will get skewed. So update Balance
 *  factors approrpiately,along the path starting from "node_in_qn" but not
 *  including him. This approach depends on the fact that,
 *  the new node is always added to a Balanced binary tree.
 *  For more details and defn of balanced binary tree,
 *  refer  "Art of Comp Programming" Donald Knuth, 3rd Volume,
 *  Sorting and Searching.
 *                                                      -- Sarnath.K
 */
        if ( node_in_qn == NULL ) {
                printk ("\nImpossible things are happening. Ok! run run\n");
                exit (1);
        }

        place = node_in_qn;
        if ( node->data < place->data )
                place = place->left;
        else
                place = place->right;

        while ( place != node ) {
                if ( node->data < place->data ) {
                        place->balance--;
                        place = place->left;
                } else {
                        place->balance++;
                        place = place->right;
                }
        }

/*
 * Now, "node_in_qn" points to the head of a minimal tree which was already
 * skewed. Now the addition of the new node could have added to the skew
 * OR balanced the skew.
 */

        dir = ( node->data < node_in_qn->data   ) ? LEFT_SKEWED : RIGHT_SKEWED;
        switch ( b(node_in_qn)  ) {
                case BALANCED:
/*
 * We added a new node to a balanced one. The tree has grown now.
 * Just leave it like that. Change the skew according to which
 * side, we added.
 */
                        b(node_in_qn) = dir; /* I think, node_in_qn == root */
                        break;

                case LEFT_SKEWED:
                        if ( dir == RIGHT_SKEWED ) {
/*
 * Got More Balanced. Honour the permutation. Lets return.
 */
                                b(node_in_qn) = BALANCED;
                                break;
                        } else {
/*
 *      The SUB-TREE with ROOT at "node_in_qn", was already skewed left.
 *      Now we have skewed it still more, thereby making him unbalanced.
 *      So, make him balanced now. The situation is Like This.

        In the following diagram A => "node_in_qn", Corresponds to the
        Scenario before adding.

        Scenario: The new node "N" is being added to the LeftSubTree of A.


                              (A)       node_in_qn
                               |
                 --------------------------
                |                          |
                L.S.A                     R.S.A
             (Skewed more)
                                ||
                                || ( Elaborating the Picture )
                                \/

                               (A)      node_in_qn
                                |
                          ---------------
                         |              |
                        (X)             R.S.A ( Ht -> H )
                         |
                   -----------
                L.S.X      R.S.X ( Ht -> H )

The CLAIM :  The Significance of "node_in_qn"
=========
        Let the node to be added = 'N'. In the above given diagram,
X and ALL its SUB_TREES should be SYMMETRICALLY BALANCED. Coz, "A"
is the nearest guy to the node to be added, who is skewed.
If X or any of it's subtrees are imbalanced (or) skewed then A
cannot be the closest node to N. X is always SYMMETICALLY BALANCED
and is of hieght H (which is equal to the hieght of Right Subtree of A). This
is Coz, if X is not SYMMETRICAL, then X would be the "node_in_qn" rather than
"A". And X's subtrees have to be of height H so that A's
         ( left subtree hieght = A's Right Subtrree hieght + 1. )
This is coz A is LEFT_SKEWED.

Now, there arise two Cases.
        The new node 'N' could have been added to any of the sub-trees
of X.
                (i)     Either  Left-Subtree of X
                (ii)    or      Right-Subtree of X
Solution
-------
                for case (i)  Rotate Right A.  Thats it.
                for case (ii) Rotate left X and Rotate Right A.

                Am i so garrulous ?? Ok, Lets return to code.
*/
                        X = node_in_qn->left;
                        if ( node->data < X->data ) { /* case (i) */
                                node_in_qn->balance = 0;
                                X->balance = 0;
                                rotate_right ( root, node_in_qn );
                        }
                        else {
                                node_in_qn->balance = 0;
                                X->balance = 0;
                                rotate_left ( root, X );
                                rotate_right ( root, node_in_qn );
                        }
                /* Oh no ! (Comment Size = n * CodeSize), n > 6  :) */
                        break;
                }

                case RIGHT_SKEWED:
                        if ( dir == LEFT_SKEWED ) {
/*
 * Got More Balanced. Honour the input sequence. Lets return.
 */
                                b(node_in_qn) = BALANCED;
                                break;
                        } else {
/*
 *
The Scenario:
-------------
                Node 'N' added to a right sub-tree of node_in_qn, which
is already skewed to the right.To know the significance of "node_in_qn", read
the big one above ( The CLAIM part ).

                                (A)  node_in_qn
                                 |
                        -----------------
                        |               |
( Height = H )        A.L.S           A.R.S ( Height = H + 1 )

                                 || ( Elaborating )
                                 ||
                                 \/

                                (A)   node_in_qn
                                 |
                        -------------------
                        |                 |
( Height H )          A.L.S               X
                                          |
                                 ----------------------
                                |                      |
                              L.S.X                  R.S.X
                            (Height H)              (Height H)

Again, Two cases arise.
                (i)  Node 'N' added to the Right Subtree of X.
                (ii) Node 'N' added to the Left Subtree of X.

Solutions:
                (i)  Rotate Left A. So that X comes to place of A.
                (ii) Rotate Right X, and Rotate left A.
 *
 */
                        X = node_in_qn->right;
                        if ( node->data >= X->data ) { /* case  (i) */
                                node_in_qn->balance = 0;
                                X->balance = 0;
                                rotate_left ( root, node_in_qn );
                        } else {        /* case (ii) */
                                node_in_qn->balance = 0;
                                X->balance = 0;
                                rotate_right ( root , X );
                                rotate_left ( root, node_in_qn );
                        }
                }
        }
        return 0;
}

/*
 * LRD - Left, Right, Data
 */
int
traverse_post_order ( struct AVLnode * node )
{
        if ( node == NULL )
                return 0;

        traverse_post_order( node->left );
        traverse_post_order( node->right );
        printk ("\n\tData -> %d, Balance -> %d", node->data, b(node));
        return 0;
}

/*
 * LDR - Left, Data, Right
 */
int
traverse_in_order ( struct AVLnode * node )
{
        if ( node == NULL )
                return 0;
        traverse_in_order( node->left );
        printk ("\n\tdata -> %d , balance -> %d", node->data, b(node));
        traverse_in_order( node->right );
        return 0;
}

/*
 * DLR ->  Data, Left, Right
 */
int
traverse_pre_order ( struct AVLnode *node )
{

        if ( node == NULL )
                return 0;

        printk ("\n\tData %d, skew %d", node->data, node->balance);
        traverse_pre_order( node->left );
        traverse_pre_order( node->right );
        return 0;
}

/*
 * Binary Search
 */
struct AVLnode *
find( long val, struct AVLnode *root )
{
	struct AVLnode *node;

	ASSERT( root != NULL );
	node = root;

	while ( node ) {
		if ( node->data == val )
			return node;
		if ( val < node->data )
			node = node->left;
		else
			node = node->right;
	}
	return NULL;
}

/*
 * Rotate the tree right, centre of rotation = node
 *
 *              Functionality.
 *              -------------
 *  0. I refer to node as 'X'.
 *
 *  1. Make sure , X has a left child. if not, u cant rotate.
 *
 *  2. Let me call X's left child as Y.

	After rotation,

 *  3. X's parent now points to Y
 *
 *  4. Make X's Left Sub-tree = Y's Right Subtree. Change the Parent of the
 *     right subtree to Point to X.

 *  5. Y's right child = X
 *
 *  6. Thats it ! EnSoy !
 */

extern int errno;

int
rotate_right (struct AVLnode **root, struct AVLnode *node)
{
        struct AVLnode *x, *y, **parptr = NULL;
        struct AVLnode *yright;

        if ( node == NULL ) {
                printk ("\n rotate_right:  NULL argument");
                return -1;
        }

        x = node;
        y = node->left;

        if ( y == NULL ) {
                printk ("\n rotate_right: attempted,  with left child NULL for \
			    node %x, node->val = %d (dec) \n", node , node->data);
                return -1;
        }

        if ( p(x) )
                parptr = (p(x)->left == x) ? &(p(x)->left) : &(p(x)->right);

        x->left = y->right;
        if ( x->left )
                x->left->parent = x;
        y->right = x;
        y->parent = x->parent;
        x->parent = y;
        if ( parptr )           // Root doesnt have parent. take care
                *parptr = y;
        else
        {
                *root = y;      // Change the root of the tree, if necessary.
        }

}

/*
 *      Functionality : Rotate a tree towards left, with centre = node.
 *      =============
 *      1. My Right Child becomes my Parent
 *      2. His LeftSubTree becomes my RightSubTree
 *      Thazz it.
 */

int
rotate_left ( struct AVLnode **root, struct AVLnode *node )
{
        struct AVLnode *x, *y, **parptr = NULL;

        if ( node == NULL ) {
                printk ("\n rotate_left: node = NULL");
                return -1;
        }

        x = node;
        y = node->right;

        if ( y == NULL ) {
                printk ("\n rotate_left: node->right = NULL");
                return -1;
        }
        /* Now start to rotate */
        if ( p(x) )
                parptr = p(x)->left == x ?  &(p(x)->left) : &(p(x)->right);

        x->right = y->left;
        if ( x->right )
                x->right->parent = x;
        y->parent = x->parent;
        y->left = x;
        x->parent = y;

        if ( parptr )
                *parptr = y;
        else
                *root = y;
}


struct AVLnode * findRightMost( struct AVLnode *node )
{
	if ( node == NULL )
		return NULL;

	while ( node->right != NULL )
		node = node->right;
	return node;
}

/*
 * AVLdelete
 * ---------
 * Functionality( Arguments: Node, Root )
 * **************************************
 *
 * 1. Find the R guy in the left subtree of the deleted node.
 *    This node represents the left neighbour of the deleted node in the input
 *    sequence.

 * 2. If this Rnode is NULL, jump to step ---

 * 3. Replace the deleted node with the R node, found in step 1.

 * 4. Let "P" point to the parent of the R node that was deleted in step 1.
 *    "P" has become lame as his right hand was mutilated. So certainly his 
 *    balance has been disrupted.

 * 5. Let "R" represent the R node that we found out in step 1.

 * 6. while( P ) {
	 if ( P->data > R->data ) {
		left = 1;
		P->balance++;
	} else {
		left = 0;
		P->balance--;
	}

	switch( P->balance ) 
	{
		case BALANCED:
		 *
	 	 *	This "node" has become more balanced ! Good !
		 *	But So bad ! One of his ancestors who was
		 *	relying on this guy's imabalance to maintain
	  	 *		his balance would have got disrupted now.
		 *	So this is not time to rest. Just continue !
		 *
			break;
	
		case RIGHT_SKEWED:
			if ( left ) {
		 *
		 *	A node in the left subtree was deleted. This is guy 
	         *	has become right skewed. which means previously he was 
		 *	neutral. This guy still maitains the balance of his 
		 * 	ancestors because his right subtree is still the same. 
		 *	By the virtue of the height of his right subtree, this 
		 *	guy can support his ancestor's balance. The Ripple ends
		 *	 here !!! Lets return from the procedure !
		 		return 0;
			} else {
		 *
		 *	A node in the right subtree was deleted. This guy got 
		 * 	right skewed. This implies => Previously this guy was 
		 *	RIGHT_IMABALANCED. This is a contradiction because the 
		 *	tree in which we are operating is a balanced binary tree
		 *
		 		print "Impossible condition !"
				panic and freeze the system !
				break;
			}
			break;

		case LEFT_SKEWED:
			if ( left ) {
		 * 
		 *	This node got left skewed after deleting a node in its 
		 *	left subtree. This implies => Previously this node's 
		 *	balance was LEFT_IMBALANCED. This is a contradiction 
		 *	because the tree in which we are operating is a 
		 *	balanced binary tree.
		 *
		 		print "Impossible condition !"
				panic and freeze the system !
				break;
			} else {
		 *
		 *	This node got left skewed after deleting a node in 
		 *	it's right subtree. This implies => prev balance 
		 *	was BALANCED. However this guy still maintains his 
		 *	ancestors balance by the virtue of his left subtree
		 *	which remains intact. Thus the ripple ends here !!! 
		 *	Lets return now.
		 *
		 		return 0;
			}

		case LEFT_UNBALANCED:
			if ( left ) {
		 		print "Impossible condition !"	
			        * Reason obvious *
				panic and freeze the system !
			} else {

		* 	Previously the node was left skewed. Now it has grown 
		*	further towards its left. The ripple doesnt end here. 
		*	The node's ancestors might have lost balance coz we 
		*	balance here. ( balancing means reduction in height )

				balance(root, P);
				if ( balancing maintained the height ) then return
			}

		case RIGHT_UNBALANCED:
			if ( left ) {
		*	Previously the node was RIGHT_SKEWED. Now he has 
		*	become right unbalanced. So Rotate left him and 
		*	continue.

				balance(root, P);
				if ( balancing maintained the height ) then return
			
			} else {
		 		print "Impossible condition !"	
			   	* Reason obvious *
				panic and freeze the system !
			}
			break;
	}
	P = P->parent;
      }
 */

int
AVLdelete( struct AVLnode *node , struct AVLnode **root )
{
	struct AVLnode *R, *P, *next;
	int left = 0, runrun = 1;

	if ( node == NULL ) {
		printk ("\navl_delete: Argument is NULL");
		return -1;
	}

#ifdef DEBUG_AVL
	called("node->data = 0x%x", node->data);
#endif

	R = findRightMost( node->left );
	if ( R == NULL ) {
		goto rightNull;
	}
	P = R->parent;

	if ( P != node )
		P->right = R->left;
	else {
		called(" P == node");
		P->left = R->left;
	}

	if ( R->left ) {
		R->left->parent = P;
	}

balance_tree:
	while ( P && runrun ) {
		next = P->parent;
		if ( P->data > R->data ) {
			left = 1;
			P->balance++;
		} else {
			left = 0;
			P->balance--;
		}

		switch(  b(P) ) {
			case BALANCED:
				break;

			case RIGHT_SKEWED:
				if ( left ) {
					goto free_and_return;
				} else {
					printk("\n Impossible Condition Hit. \
						   Halting");
					while(1);
				}
				break;

			case LEFT_SKEWED:
				if ( left ) {
					printk("\n Impossible Condition Hit. \
						   Halting");
					while(1);
				} else {
					goto free_and_return;
				}
				break;

			case LEFT_UNBALANCED:
				if ( left ) {
					printk("\n Impossible Condition Met");
					while(1);
				} else {
					runrun = 
				  (balance(P, root) == HEIGHT_REDUCED) ?  1 : 0;
				}
				break;

			case RIGHT_UNBALANCED:
				if ( left ) {
					runrun = 
				  (balance(P, root) == HEIGHT_REDUCED) ?  1 : 0;
				} else {
					printk("\n Impossible condition met");
					printk("\n Aborting");
					while(1);
				}
		}
		P = next;
	}

free_and_return:
	if ( R ) {
		node->data = R->data;
		free(R);
	}
	return 0;

rightNull:
#ifdef DEBUG_AVL
	called("Rightmost is NULL !!");
#endif
	/*
	 * mm..I am tired of handling exceptions.
	 * No escape. get ....ed up !
 	 * We came here because node->left is NULL
 	 */
	if ( node->right ) {
	/* Since node->left is NULL, node->right will not have any chilren ! 
	 * Why ?? Coz, this is a balanced tree.
	 */
		//node->data = node->right->data;
		R = node->right;
		node->right = NULL;
		P = node;
		goto balance_tree;
	}
	else {
		/* A leaf is being deleted */
		#ifdef DEBUG_AVL
			called(" Leaf deleted !!");
		#endif
		R = node;
		P = node->parent;
		if ( P == NULL ) {
			*root = NULL;
		#ifdef DEBUG_AVL
			called(" Root deleted !!");
		#endif
			goto free_and_return;
		} else {
			if ( P->left == R )
				P->left = NULL;
			else
				P->right = NULL;
			goto balance_tree;
		}
	}
	/* 
	 * We shall never reach here.
	 */
	printk("\n This code shall not be executed");
	return -1;
}

/*
 * Generic function. - Shall balance a minimally unbalanced tree.
 * minimally unbalance - b(P) = -2 to 2. Not more than that.
 *
 * Return Values:
 * 	-1 -> Wrongly Called.
 *	 0 -> Balanced and Height remained the same.
 *	 1 -> Balanced and Height reduced.
 */

int
balance( struct AVLnode *P, struct AVLnode **root )
{
	struct AVLnode *X, *Y;
	int retval;

	retval = HEIGHT_REDUCED;

	if ( b(P) == LEFT_UNBALANCED ) {
		X = P->left;
		switch ( b(X) ) {
			case BALANCED:
				rotate_right( root, P);
				b(X) = RIGHT_SKEWED;
				b(P) = LEFT_SKEWED;
				retval = HEIGHT_MAINTAINED;
				break;
			case LEFT_SKEWED:
				rotate_right( root, P );
				b(X) = BALANCED;
				b(P) = BALANCED;
				break;
			case RIGHT_SKEWED:
			/*
			 * In this case, rotate_right doesn't solve the problem.
			 */
				Y = X->right;
				switch( b(Y) ) {
					case LEFT_SKEWED:
						b(X) = BALANCED;
						b(Y) = LEFT_UNBALANCED;
						break;
					case RIGHT_SKEWED:
						b(Y) = b(X) = LEFT_SKEWED;
						break;
					case BALANCED:
						b(X) = BALANCED;
						b(Y) = LEFT_SKEWED;
						break;
				}
				rotate_left(  root, X  );
				switch( b(Y) ) {
					case LEFT_SKEWED:
						b(P) = b(Y) = BALANCED;
						break;
					case LEFT_UNBALANCED:
						b(P) = RIGHT_SKEWED;
						b(Y) = BALANCED;
						break;
				}
				rotate_right( root,P );
				break;		
		}
		return retval;
	}

	if (b(P) == RIGHT_UNBALANCED) {
		X = P->right;
		switch ( b(X) ) {
			case BALANCED:
				rotate_left(root,P);
				b(X)   = LEFT_SKEWED;
				b(P)   = RIGHT_SKEWED;
				retval = HEIGHT_MAINTAINED;
				break;
			case RIGHT_SKEWED:
				rotate_left(root,P);
				b(X) = BALANCED;
				b(P) = BALANCED;
				break;
			case LEFT_SKEWED:
			/*
			 * In this case, rotate_left doesnt solve the problem.
			 */
				Y = X->left;
				switch( b(Y) ) {
					case RIGHT_SKEWED:
						b(X) = BALANCED;
						b(Y) = RIGHT_UNBALANCED;
						break;
					case LEFT_SKEWED:
						b(Y) = b(X) = RIGHT_SKEWED;
						break;
					case BALANCED:
						b(X) = BALANCED;
						b(Y) = RIGHT_SKEWED;
						break;
				}
				rotate_right( root, X );

				switch( b(Y) ) {
					case RIGHT_SKEWED:
						b(P) = b(Y) = BALANCED;
						break;
					case RIGHT_UNBALANCED:
						b(P) = LEFT_SKEWED;
						b(Y) = BALANCED;
						break;
				}
				rotate_left( root, P );
				break;
		}
		return retval; 
	}
	called("\n Lotta wrong things happening on earth");
	return ILLEGAL;
}
