/*
 * Set TabStop to 4 for the correct view.
 *
 * ABOUT:
 *
 * 	An O(1) memory allocator for kernel memory. Coalescing of free-fragments is
 *  done in an O(1) fashion unlike other allocators.
 *
 */

#include <malloc.h>
#include <sys/time.h>

/**** Macros and Data Structures required for the main memory allocator ***/
/* { */

#undef MALLOC_DEBUG
#define MALLOC_SIGNATURE	0x0DEADBED
#define SUCCESS				0
#define	FAILURE				!SUCCESS

#define RESOURCE(x)	(((unsigned long)resource) + (unsigned long)(x))
#define OFFSET(x)	((unsigned long)(x) - (unsigned long)(resource))

#define ASSERT(x) \
		if ((x) == 0) { \
			printf("\n Assertion failed "); \
			while(1);\
		}

#define GLOBAL_LIST_UPDATE(ptr, which, what)\
		if (*((unsigned long *)ptr) == MALLOC_SIGNATURE)\
		{\
			struct free_frag *f_macro;\
\
			f_macro = (struct free_frag *)ptr;\
			f_macro->which = what; \
		} else {\
			struct o1suffix *sunique_macro;\
\
			sunique_macro = (struct o1suffix *)ptr;\
			sunique_macro->which = what;	\
		}

#define FREE(x) ((*((unsigned long *)x) == MALLOC_SIGNATURE) ? 1 : 0)

#define DUMP_FREE_FRAG(fr)\
{\
	struct free_frag *f_macro;\
\
	f_macro=(struct free_frag *)fr;\
	printf("\n DUMP of free fragment 0x%x", f_macro);\
	printf("\nSignature = 0x%x", f_macro->signature);\
	printf("\nStart = 0x%x", f_macro->start);\
	printf("\nsize = 0x%x", f_macro->size);\
	printf("\nprev = 0x%x", f_macro->prev);\
	printf("\nnext = 0x%x", f_macro->next);\
	printf("\nnext_free = 0x%x", f_macro->next_free);\
	printf("\nprev_free = 0x%x", f_macro->prev_free);\
	printf("\nindex = 0x%x", f_macro->index);\
}

#define DUMP_O1SUFFIX(f)	\
{\
	struct o1suffix *sunique_macro;\
\
	sunique_macro=(struct o1suffix *)f;\
	printf("\n Dump of O1Suffix 0x%x", sunique_macro);\
	printf("\n Next = 0x%x", sunique_macro->next);\
	printf("\n prev = 0x%x", sunique_macro->prev );\
}

#define DUMP_FRAG(f)\
	if (*(unsigned long *)f  == MALLOC_SIGNATURE)\
	{\
		DUMP_FREE_FRAG(f)\
	} else {\
		DUMP_O1SUFFIX(f)\
	}

#define NULL 0

struct free_frag
{
	unsigned long signature;	/* This has to be the first field. */
	unsigned long start, size;
	void *prev, *next;
	struct free_frag *next_free, *prev_free; 
	int index; /* "index" of the bucket */
};

/*
 *	Data Structure: struct free_frag
 ***********************************
 *
 *	prev,next:
 *	Linked list of all fragments - in ascending order of addresses. 
 *	IRRESPECTIVE of whether they are free or allocated. The pointers are of 
 *	type "free_frag" if the pointer points to private memory. Otherwise it is 
 *	of type "o1suffix". Ordering is maintained by the very nature of memory 
 *	allocation. No explicit sorting is done. It is this aspect that makes 
 *	coalescing an ORDER operation.
 * 
 *	next_free, prev_free:
 *	Linked list of free fragments for this size. Unordered list.
 *
 */

struct free_frag_list
{
	unsigned long size;
	struct free_frag *free_list_head;
	struct free_frag *free_list_tail;
};

/*
 *	Data Structure: free_frag_list (bucket entry)
 ********************************
 *	This data structure forms the entry in each bucket. It contains the linked 
 *	list of free-fragments (in no specific order) for that particular size.
 */

struct o1suffix {
	void *next, *prev;
};

/*
 *	Data Structure: o1suffix
 ***************************
 *	This structure is appened below each allocated memory block. It helps to
 *	maintain a linked list of all fragments in the order of their addresses.
 *	It is this suffix that gives an O(1) coalescing power to this algorithm.
 */

#define NUM_BUCKETS		26
#define WASTE_BUCKET	(NUM_BUCKETS+1)
#define MIN_SIZE		16
#define LOG_MIN_SIZE	4
#define THRESHOLD_SIZE	((unsigned long)(1 << ((NUM_BUCKETS-1)+LOG_MIN_SIZE)))

struct free_frag_list buckets[NUM_BUCKETS + 1];
/*
 *	The extra bucket is for the waste bucket. This bucket holds free fragments
 *  whose sizes are less than MIN_SIZE. They are not considered for any
 *	allocation. They reside in this bucket untill they are merged with their
 *	neighbours.
 */

/*** End Data Structures required for the main memory allocator *********?
/* } */

static void malloc_init();
static void * allocate_private_memory();
static void free_private_memory(void *freed);
static int find_bucket_index(unsigned long size);

/******* DATA STRUCTURES AND ROUTINES FOR INITIALIZATION AND HANDLING 
         PRIVATE-MEMORY ************/
/* { */

#define MAX_FREE_FRAGMENTS 512
struct private_mem_internals {
	struct free_frag data;
	struct private_mem_internals *next_free;
};
struct private_mem_internals *private_memory, *private_free_head;

#define RESOURCE_SIZE	( 1 << 28 )	
void *resource;

int malloc_init_happened = 0;

static void
malloc_init()
{
	int i,index;
	struct free_frag *temp;

	for(i=0; i< NUM_BUCKETS; i++)
	{
		buckets[i].size = (1 << (i+LOG_MIN_SIZE));
		buckets[i].free_list_head = NULL;
		buckets[i].free_list_tail = NULL;
	}

	buckets[WASTE_BUCKET-1].size = 0;
	buckets[WASTE_BUCKET-1].free_list_head = NULL;
	buckets[WASTE_BUCKET-1].free_list_tail = NULL;
/*
 *	The waste bucket holds free-fragments which are less than MIN_SIZE. They 
 *	will not be considered for allocatins. These fragments sit there until they
 *	get coalesced with their neighbours.
 */

	private_memory = (struct private_mem_internals *) 
	malloc(MAX_FREE_FRAGMENTS * sizeof(struct private_mem_internals));

	if (private_memory == NULL)
	{
		perror("\nUnable to allocate private mem: ");
		exit(1);
	}

	memset((void *)private_memory,0,
	MAX_FREE_FRAGMENTS*sizeof(struct private_mem_internals));

	for(i=0; i<MAX_FREE_FRAGMENTS-1; i++)
	{
		private_memory[i].next_free = &private_memory[i+1];
		private_memory[i].data.signature = MALLOC_SIGNATURE;
	}
	private_memory[i].next_free = NULL;
	private_free_head = &(private_memory[0]);

	resource = malloc(RESOURCE_SIZE);
	if (resource == NULL)
	{
		perror("\nUnable to allocate memory for the resource");
		exit(1);
	}

	/* Allocate the Master free fragment - The parent of the Big Bang. */

	temp = (struct free_frag *) allocate_private_memory();
	index= find_bucket_index(RESOURCE_SIZE);
	buckets[index].free_list_head = temp;
	buckets[index].free_list_tail = temp;
	if (temp == NULL)
	{
		perror("\n First allocation to private memory itself failed");
		exit(1);
	}
	
	temp->start = 0;
	temp->size =  RESOURCE_SIZE;
	temp->prev = NULL;
	temp->next = NULL;
	temp->next_free = NULL;
	temp->prev_free = NULL;
	temp->index = index;

	malloc_init_happened=1;
	return;
}

static void *
allocate_private_memory()
{
	void *temp;

	if (private_free_head)
 	{
		temp = (void *)private_free_head;
		private_free_head = private_free_head->next_free;
		return temp;
	}
	return NULL;
}

static void
free_private_memory(void *freed)
{
	struct private_mem_internals *temp;

	temp = (struct private_mem_internals *)freed;

	if (private_free_head == NULL)
	{
		private_free_head = temp;
		temp->next_free = NULL;
		return;
	}
	
	temp->next_free = private_free_head;
	private_free_head = temp;
	return;
}

/*********** End INIT and Private Mem Handling Routines *********************/
/* } */


/********* Support Functions and the Main memory allocator *****************/
/* { */

/**************************************************************************
 * function: find_bucket_index(unsigned long size)
 *
 * functionality:
 * The function uses binary search to find the appropriate bucket for the 
 * requested size. For example, it returns 128th bucket index for any
 * size between 128 to 255. The time-complexity being: O(log(NUM_BUCKETS)); 
 * Never more than 5 for 32-buckets.
 *
 **************************************************************************/
static int 
find_bucket_index(unsigned long size)
{
	int lo,hi;

	lo=NUM_BUCKETS/2;
	hi=NUM_BUCKETS-1;

	while(1)
	{
		if (size == buckets[lo].size)
			return lo;

		if (size > buckets[lo].size)
		{
			int temp;

			if (lo == (NUM_BUCKETS-1))
				return lo;
			
			temp = lo;
			lo=(lo+hi)/2;
			if ( lo == temp) 
			{
				lo=hi;
			}
			continue;
		}

		if (lo != 0)
		{
			if (size > buckets[lo-1].size) {
				return lo-1;
			}

			hi=lo;
			lo=lo/2;
		}
		else
		{
			printf("\n We should never get his condition. The given size");
			printf(" should be greater than MIN_SIZE");
			printf("\n The given size = 0x%x", size);
			return lo;
		}
	}
}

/*****************************************************************************
 * Function: static struct free_frag * find_frag(struct free_frag *p, 
 *												 unsigned long size)
 *
 * Functionality: 
 * This function searches a linked list (un-ordered) for a fragment of size
 * >= requested size. Time complexity = O(length of list). However this search 
 * can be avoided by increasing NUM_BUCKETS so that we handle sizes >= 
 * RESOURCE_SIZE in an O(1) way without calling this function. If we use lesser 
 * NUM_BUCKETS then this code will be hit for larger sizes. The NUM_BUCKETS is 
 * a tuning parameter for this allocator. For a 4GB RESOURCE_SIZE and 
 * NUM_BUCKETS of 28 this code will never need to be called. So dont jump into 
 * conclusion that the allocator is in-efficient after seeing the time
 * complexity associated with this allocator.
 *****************************************************************************/ 
static struct free_frag *
find_frag(struct free_frag *p, unsigned long size)
{
	while (p)
	{
		if (p->size >= size)
			return p;
		p = p->next_free;
	}
	return NULL;
}

/******************************************************************************
 * Function:	static void detach_from_free_list(struct free_frag *frag);
 * 
 * Functionality:
 * As expected, this function just deletes a free fragment from the list.
 * Involves basic doubly-linked list deletion code. 
 *****************************************************************************/
static void
detach_from_free_list(struct free_frag *frag)
{
	struct free_frag_list *f;

	f = &buckets[frag->index];

	if (f->free_list_head == frag)
		f->free_list_head = frag->next_free;
	if (f->free_list_tail == frag)
		f->free_list_tail = frag->prev_free;

	if (frag->prev_free)
	{
		frag->prev_free->next_free = frag->next_free;
	}

	if(frag->next_free)
	{
		frag->next_free->prev_free = frag->prev_free;
	}

	frag->next_free = NULL;
	frag->prev_free = NULL;
	return;
}

/*****************************************************************************
 * Function: enqueue_in_free_list(struct free_frag *frag, int index)
 *
 * Functionality:
 * As expected, this function moves a fragment from one list to another.
 * The caller specifies the "index" of the list to which the fragment should
 * be enqueued. The code just adds the fragment to the tail of the list.
 * Involves mere doubly-linked list addition.
 *
 *****************************************************************************/
static void
enqueue_in_free_list(struct free_frag *frag, int index)
{
	struct free_frag_list *f = &buckets[index];

	frag->index = index;

	if ((f->free_list_tail == NULL) || (f->free_list_head == NULL))
	{
		f->free_list_tail = frag;
		f->free_list_head = frag;
		frag->next_free = NULL;
		frag->prev_free = NULL;
		return;
	}

	ASSERT(f->free_list_tail->next_free == NULL);
	f->free_list_tail->next_free= frag;
	frag->prev_free = f->free_list_tail;
	frag->next_free = NULL;
	f->free_list_tail = frag;
	return;
}

/*****************************************************************************
 * Function: attach_to_free_list(struct free_frag *frag, unsigned long index)
 *
 * Functionality:
 * The functionality is similar to the "enqueue" function. However this
 * function adds the fragment to the head of the list rather than the tail.
 *
 **************************************************************************/
static void
attach_to_free_list(struct free_frag *frag, unsigned long index)
{
	struct free_frag_list *f;

	f=&buckets[index];
	frag->index=index;
	if (f->free_list_head == NULL)
	{
		f->free_list_head = frag;
		f->free_list_tail = frag;
		frag->next_free = NULL;
		frag->prev_free = NULL;
		return;
	}

	frag->next_free = f->free_list_head;
	frag->prev_free = NULL;
	f->free_list_head->prev=frag;
	f->free_list_head = frag;
	return;
}

/*****************************************************************************
 * Function: break_fragment( struct free_frag_list *, unsigned long size )
 *
 * Caller: "kmalloc" calls this function to allocate memory.
 *
 * Functionality:
 * This function does the core of the functionality of "kmalloc". The function
 * breaks "size" bytes from the fragment "free_frag_list" and returns the start
 * address of the fragment.
 *
 * Assumption:
 * The caller has to make sure that the "free_frag_list *" contains fragments
 * which are greater than size "size".
 *
 *****************************************************************************/
static unsigned long
break_fragment(struct free_frag_list *p, unsigned long size)
{
	struct free_frag *frag;
	int index;
	unsigned long retval=0;
	struct o1suffix *suffix;

/*
 * First find out the fragment that can support it.
 */
	if (p->size == THRESHOLD_SIZE)
	{
		frag = find_frag(p->free_list_head,size);
		if (frag == NULL)
		{
			return NULL;
		}
	} else {
		frag = p->free_list_head;
	}

	if (frag == NULL) 
	{
/* There is still a chance of being NULL. Checkout the caller */
		return NULL;
	}

	ASSERT(frag->size >= size);

/*
 *	Now break the fragment.
 */
	retval = frag->start;
	frag->size 	-= size;
	frag->start += size;

	if (frag->size != 0)
	{
		/*
		 *	Now that the fragment has shrinked, move it to an approriate
		 *	position if necessary.
		 */
		if (frag->size < MIN_SIZE)
		{
			index = WASTE_BUCKET;
		} else {
			index = find_bucket_index(frag->size);
		}

		if (index != frag->index)
		{
			detach_from_free_list(frag);
			enqueue_in_free_list(frag, index);
		}

		/*
		 *	Update the Global List of fragments. The updation maintains the
		 *	order of addresses.
		 */
		suffix = (struct o1suffix *)( ((unsigned long)resource) + retval +
										size - sizeof(struct o1suffix));
		suffix->next = frag;
		suffix->prev = frag->prev;
		if (frag->prev)
		{
			GLOBAL_LIST_UPDATE(frag->prev, next, suffix);
		}
		frag->prev = suffix;
	} else {
		/*
		 * The fragment has been completely exhausted. Update the Global
		 * list of fragments. This updation maintains the order of addresses.
		 */
		suffix = (struct o1suffix *)( RESOURCE(retval) + size - 
									  sizeof(struct o1suffix));
		suffix->next = frag->next;
		suffix->prev = frag->prev;
		if (frag->prev)
		{
			GLOBAL_LIST_UPDATE(frag->prev, next, suffix);
		}
		if (frag->next)
		{
			GLOBAL_LIST_UPDATE(frag->next, prev, suffix);
		}
		frag->next = NULL;
		frag->prev = NULL;
		free_private_memory(frag);
	}
	return	RESOURCE(retval); 
}

/******************************************************************************
 * Function: kmalloc(unsigned long size)
 *
 * Functionality:
 * This is the interface function provided to the rest of the kernel to
 * allocate kernel memory.
 *****************************************************************************/
unsigned long
kmalloc(unsigned long size)
{
	int index, original_index;
	unsigned long retval;

	ASSERT(malloc_init_happened != 0)

	if (size < MIN_SIZE)
	{
		size = MIN_SIZE;
	}
	size += sizeof(struct o1suffix);

	if (size > RESOURCE_SIZE)
	{
		printf("\n Unable to handle BIG size");
		return NULL;
	}

	index = find_bucket_index(size);
	original_index = index;

	while( (buckets[index].free_list_head == NULL) && (index < NUM_BUCKETS))
	{
		index++;
	}

	if (index == NUM_BUCKETS)
	{
	#ifdef MALLOC_DEBUG
		printf("\nkmalloc: No free memory available.");
	#endif
		return NULL;
	}

	if ((original_index == index) && ((index+1) != NUM_BUCKETS))
	{
		index++;
	}

	retval = break_fragment(&buckets[index], size);
	return retval;
}

/****************************************************************************
 * Function: coalescable(void *ptr, unsigned long size) 
 *
 * Functionality:
 * The function returns SUCCESS if the freed kernel memory can be coalesced
 * with any of its neighbouring chunk. Otherwise it returns FAILURE.
 ***************************************************************************/
static inline int
coalescable(void *ptr, unsigned long size)
{
	struct o1suffix *s;

	s=(struct o1suffix *)((char *)ptr+size);
	if (((s->next != NULL) && FREE(s->next)) || ((s->prev != NULL) && FREE(s->prev)))
		return SUCCESS;
	return FAILURE;
}

/******************************************************************************
 * Function: static void coalesce(void *ptr, unsigned long size)
 *
 * Functionality:
 * The argument passed is the freed memory pointer and freed size. There can
 * be a maximum of only 2 coalescing operations possible. This is because:
 *	   a) At any point of time the global-list is a condensed list. i.e no
 *	   coalecing is possible on the list unless a FREE happens.
 *	   b) The introduction of a new Free Chunk F2 between F1 and F3 can cause
 *		  a coalescing of F1-F2-F3 into 1 chunk. If F3 of F1 is not a free
 *		  fragment then we can have only 1 coalesce operation. i.e F1-F2 or
 *		  F2-F3.
 * Notion:
 * The following notion is followed in the comments:
 *	F2 ->  Refers to the fragment being freed.
 *	F1 ->  F2's Previous Fragment.
 *  F3 ->  F2's Next Fragment.
 *****************************************************************************/
static void
coalesce(void *ptr, unsigned long size)
{
	struct o1suffix *s;
	struct free_frag *f1, *f3, *affected;
	char f1_coalesced = 0, f3_coalesced = 0;
	int index;

	s = (struct o1suffix *)((char *)ptr + size);
	f1 = s->prev;
	f3 = s->next;

	#ifdef MALLOC_DEBUG
	printf("\n f1 = 0x%x, f3=0x%x", f1, f3);
	#endif

	if (f1 != NULL)
	{
		if (FREE(f1))
		{
			#ifdef MALLOC_DEBUG
			printf("\n A colescable F1 found");
			#endif
			/* Coalesce F1 and F2 */
			f1_coalesced = 1;
			f1->size += (size + sizeof(struct o1suffix));
			f1->next = s->next;
			if (s->next)
			{
				GLOBAL_LIST_UPDATE(s->next, prev, f1);
			}
		}
	}

	if (f3 != NULL)
	{
		if (FREE(f3))
		{
			#ifdef MALLOC_DEBUG
			printf("\ncoalesce: Free F3 Fragment found");
			#endif
			if (f1_coalesced)
			{
				/* Now F1 and F3 had to be coalesced. */
				f1->size += (f3->size);
	            f1->next = f3->next;
	            if (f3->next)
	            {
	                GLOBAL_LIST_UPDATE(f3->next, prev, f1);
	            }
				free_private_memory((void *)f3);
			} else {
			   /* Coalesce F2 and F3 */
				f3->size += (size + sizeof(struct o1suffix));
				f3->start = OFFSET(ptr);
				if (s->prev)
				{
					GLOBAL_LIST_UPDATE(s->prev, next, f3);
				}
				f3->prev = s->prev;
				f3_coalesced = 1;
			}
		}
	}

	/*
	 * Now either F1 or F3 or none has changed in size. So fit them in the
	 * correct bucket.
	 */
	if (f1_coalesced)
	{
		affected = f1;
	}
	else if (f3_coalesced)
	{
		affected = f3;
	}
	
	if (affected == NULL)
	{
		return;
	}
	
	index = find_bucket_index(affected->size);
	if (index != affected->index)
	{
		detach_from_free_list(affected);
		#ifdef MALLOC_DEBUG
		printf("\nAffected fragment moved from index %d to index %d", 
				affected->index,index);
		#endif
		attach_to_free_list(affected, index);	
	}
	return;
}

/******************************************************************************
 * Function: kfree(void *ptr, unsigned long size)
 *
 * Functionality:
 * This function is an interface function which is exported to the rest of the
 * kernel. The function also takes care of coalescing the chunk with its 
 * neighbours.
 *****************************************************************************/
int
kfree(void *ptr, unsigned long size)
{
	struct free_frag *host;
	struct o1suffix *s;
	unsigned long index;

	ASSERT(size <= RESOURCE_SIZE)
	if (size < MIN_SIZE)
	{
		size=MIN_SIZE;
	}

	s=(struct o1suffix *)( ((unsigned long)ptr)+size);

	if (coalescable(ptr,size) != SUCCESS)
	{
		#ifdef MALLOC_DEBUG
		printf("\nkfree: No Coalescable fragment found.");
		#endif
		host= (struct free_frag *)allocate_private_memory();
		if (host == NULL)
		{
			printf("\nkfree: Unable to allocate internal memory");
			return -1;
		}

		ASSERT(host->signature == MALLOC_SIGNATURE)
		host->start = OFFSET(ptr);
		host->size = size+sizeof(struct o1suffix);

		/* Update the Global List appropriately */
		if (s->prev)
		{
			GLOBAL_LIST_UPDATE(s->prev,next,host);
		}

		if (s->next)
		{
			GLOBAL_LIST_UPDATE(s->next,prev,host);
		}
		host->next=s->next;
		host->prev=s->prev;

		/* Insert the free fragment into an appropriate list. */
		index=find_bucket_index(host->size);
		attach_to_free_list(host,index);
		return SUCCESS;
	} else {
		#ifdef MALLOC_DEBUG
		printf("\nkfree: coalescable fragment found. Proceeding to coalesce");
		#endif
		coalesce(ptr,size);
	}
	return 0;
}


/******** END Support Functions and the Main memory allocator ***************/
/* } */

char *ptr[300];

struct t 
{
	int a[2];
};

int
main()
{
	struct t *a, *b;
	struct o1suffix *x;
	struct timeval start,end;
	unsigned long size;
	int i;

	malloc_init();

	a=(struct t *)kmalloc(sizeof(struct t));
	memset((void *)a,0,sizeof(struct t));
	b=(struct t *)kmalloc(sizeof(struct t));

	// memset((void *)a,0, sizeof(struct t));	
	printf("\n a = 0x%x, b=0x%x \n", a, b);
	return 0;
}
