/*
 *                      Generic Binary Tree Interface for Linux.
 * Assumption:
 * Any structure with a binary tree arrangement must have 3 fields
 * 1. left
 * 2. right
 * 3. parent
 */

#ifndef __LIN_BINTREE_H__
#define  __LIN_BINTREE_H__

#ifndef NULL
#define NULL 0
#endif

#define OFFSET(member,structure) (unsigned int)(&((struct structure *)0)->member)
#define LEFT(ptr, cred) ((void **)((char *)ptr + cred->left))
#define RIGHT(ptr, cred) ((void **)((char *)ptr + cred->right))
#define PARENT(ptr, cred) ((void **)((char *)ptr + cred->parent))

struct b_credentials {
        unsigned int left,right,parent;
                        /*
                         * left,right,parent => offsets of the "left" , "right"
                         * , "parent" pointers in the caller's structure
                         */
        int (*compare)();
        void *minimum;  /* Left most leaf of the binary tree */
};

static int
bin_insert_subtree ( void * node, void *root, struct b_credentials *cred );

static void *
minimum ( void *root, struct b_credentials *cred )
{
        void *left;

        while ( root ) {
                if ( (left = *LEFT(root, cred)) == NULL )
                        return root;
                root = left;
        }
        return NULL;
}

                /* Interface function */
static int
bin_insert ( void * node , void ** root, struct b_credentials *cred )
{
        *LEFT(node, cred) = NULL;
        *RIGHT(node, cred) = NULL;
        *PARENT(node, cred) = NULL;

        if ( *root == NULL ) {
                *root = node;
                cred->minimum = node;
                return 0;
        }
        bin_insert_subtree ( node, *root, cred );
        if ( *LEFT(cred->minimum, cred) == node )
                cred->minimum = node;
        return 0;
}

static int
bin_insert_subtree ( void * node, void *root, struct b_credentials *cred )
{
        void *tr, *parent,  **posn;

        if ( node == NULL  || root == NULL)
                return -1;
        tr = (void *) root;
        if ( tr == NULL )
                return -1;
        while ( tr ) {
                parent = tr;
                if ( cred->compare( node, tr ) < 0 ) {
                        tr =  *(LEFT(tr, cred));
                } else {
                        tr =  *(RIGHT(tr, cred));
                }
        }
        if ( cred->compare (node, parent) < 0 )
                *LEFT(parent, cred) = node;
        else
                *RIGHT(parent, cred) = node;
        *PARENT(node, cred) = parent;
}

                /* Interface functions */
static int
bin_delete ( void *node, void **root, struct b_credentials *cred )
{
        void *right, *left, *parent, *successor, *other;
        void *sibling, **posn;

/*
 *      P -> Node to be deleted
 *      Just Make L or R, the successor of Parent and insert the remaining
 *      subtree into the successor sub-tree.

                Parent                                  Parent
                |                                        |
         --------                                        ----------
         |                                                        |
         P                                                        P
 ----------------                                       ---------------------
|               |                                       |                   |
L               R                                       L                   R
 *
 */

                if ( node == NULL || root == NULL )
                        return -1;

                right = *RIGHT(node, cred);
                left = *LEFT(node, cred);
                parent = *PARENT(node, cred);

/*
 *      First , check if we need to find minimum again.
 *      If "node" is the Left-Most-Leaf then
 *      1. "node" doesn't have a left sub-tree.
 *      2.  node is always attached to the parent via the LEFT
 *      pointer (otherwise) parent would be the minimum
 */
        if ( cred->minimum == node ) {
        #ifdef DEBUG
                ASSERT( left == NULL );
        #endif
                if ( right )
                        cred->minimum = minimum( right , cred );
                 else
                        cred->minimum = parent;
        }
/*
 *      Now, minimum is in a sane state.
 *      Start the deletion
 */
        if ( right ) {
                successor = right;
                other = left;
        } else {
                successor = left;
                other = right;
        }
        if ( parent == NULL )
                *root = successor;
        else {
                posn = (cred->compare(node, parent)) < 0 ? LEFT(parent, cred) :
                                                      RIGHT(parent, cred);
                *posn = successor;
        }
        if ( successor )
                *PARENT(successor, cred) = parent;
        if ( other && successor )
                bin_insert_subtree ( other , successor, cred );
        return 0;
}
#endif
