/*
 *  This file is supposed to have the xlation related data structures.
 *                                              -- Sarnath.K
 */

#ifndef __PTOV_H__
#define __PTOV_H__
#include <config.h>
#include <generic/types.h>
#include <i386/page.h>
#include <mm/vm.h>

/*
 * The structure "ptov" helps in identifying the various mappings that a
 * physical page has. Inorder to know about the mappings for a physical page,
 * get on  to its "pmem_t" structure, and follow the linked list of "ptov"s to
 * find out, what all {address-space, va} map the given physical address.
 * Note:
 *      Regarding memory allocation of ptov structures, i dont use
 * kmem_allocs, coz the problem becomes recursive. And moreover, i dont
 * want to waste kernel virtual addresses ( by opting for a seperate kernel
 * segment or any other. Already we have just 1 GB.)
 * Hence, i would just "pmem_alloc" a page and try accomodating "ptov"
 * structures inside the page. The PTRs in the linked list of
 * PTOV structures are also physical.
 * All my conecern is about internal allocation within this page. I do that by 
 * having meta-data about such a page, ( let me call it "unmapped page" ), at 
 * the page head itself. This should be fairly good.
 *
 */

typedef struct ptov {
        ulong va;
        as_t  *owner;           // Owning Address Space.
        ulong refcnt;           // Same VA can map more than once.
                                // For exp, e0123000 and e0123080 may be
                                // two kmem_alloced VAs and can point to same
                                // physical address coz they belong to same
                                // virtual page.
        struct ptov *next;      // doubly linked list, physical addresses
        struct ptov *prev;      // Get a temp mapping before indirecting these
                                // PTRs.
                                // else u might land up in a panic.
                                //              (or)
                                // i can have an alternate PGDIR in which
                                // address 0 to 3g are mapped linearly and
                                // 3G to 4G should b IIIr to kernel page
                                // directory. In this case, make sure while
                                // allocating phy memory for a ptov structure,
                                // it should be less than 3G physical else i
                                // cant map them in the new page-directory. If
                                // i do so, i'll lose kernel mappings since
                                // kernel resides above 3gb offset. (coz kernel
                                // code should execute after switching to the
                                // new page directory. And all kernel code have
                                // been relocated to 3Gb.) This method
                                // is very suave. The previous method
                                // will just slow down performance due to
                                // constant TLB flushes. but this method is
                                // memory expensive, coz u need 768
                                // more page tables. which = 3mb of RAM.
                                //              Bap re !
                                // I have chosen the Former method. The second
                                // method can also be implemented in a
                                // lazy way of page-table allocation. but let
                                // me play it safer. I'd adopt the 2nd method
                                // some other day if i feel this is a
                                // bottleneck.
}ptov_t;

/*
 * The Math -->
 * Let "x" number of  "ptov" structures that can be accomodated in a page.
 * "x" bits are required for managing them other than two ulongs.
 * (See struct unmapped_page_head )
 * Hence the math is = "x/8 +  8" + x*sizeof(ptov) <= PAGESIZE
 * = x/8 + 20*x <= PAGESIZE - 8
 * = 161*x/8 <= 4088
 * = x <= (4088 * 8) / 161
 * = x <=  203.13
 * = Bitfield size = 203/8 = 25.375 = Max interger = 25
 * so numslots = 8*25 = 200.
 * Total used bytes in the page = 200/8 + 8 + (200*20) =  4033.
 * Remaining 63 bytes are wasted. Fine, thats a compromise i make
 * for a simpler code.
 * 
 */

 #define NUM_PTOVS 200
// (25*8) PTOV structures can be accomodated in a
// single page along with "unamapped_page_head"

typedef struct unmapped_page_head {
        unsigned char  bitfield[NUM_PTOVS/8];
        ulong entries; // num-valid entries in the page. When it falls to 0 , i
                       // would pmem-free the page.
        struct unmapped_page_head *next;
}unmapped_page_t;

extern unmapped_page_t  *start_unmapped_page; // Pointer is a physical address.
#endif
