/*****************************************************************
/*
/*	NOTE :: Set Tab Space to 4 for alignment::
/*
/*	Copyright and Disclaimer	
/*											Sarnath		2002
/*
/*		You are free to distribute and re-use the code for
/*		any purpose except commercial purpose without hiding
/*		the author(s)'s identity. However I strongly
/*		disclaim any explicit or implicit usability of the code.
/*		I wont be responsible to any damage caused due to the
/*		study or use of this code. Thanks for your patience.
/*	
/* 	Contents
/*
/* 		This file contains routines for allocation and freeing
/*		GDT,LDT entries. It also contains the function for
/*		creating segment descriptors. All of them are 386 dependent
/*		routines.
/*                                              - Sarnath.K
/*****************************************************************/

#include <generic/types.h>
#include <threads/thread.h>
#include <i386/descr.h>
#include <i386/tss.h>

ldt_t *ldt_pool;

ulong gdtbits[ 256 ]; /* 1024/4 = 256 */
/*****************************************************************************
/*
/*	Description	:
/*		
/*	Each bit in this array indicates the availability of the corresponding
/*	GDT entry. BIT number 0 of gdtbits[0] corresponds to GDT entry 0.  
/*	There are totally 8*256*4 bits in the entire array. Thus 8192 entries 
/*	are	covered by the array.
/*
/****************************************************************************/

/*******************************************************
/*	Function	:	gdt_init
/*
/*	Callers		:	Called during kernel initzn once.
/*
/*	Description	:
/*
/*		This function allocates the LDT pool and
/*		initializes the "gdtbits" array to indicate
/*		the usage of the GDT.
/*
/*******************************************************/

ushort
gdt_init()
{
        gdtbits [0] = 0x1F;
/*--------------------------------------------
/*      1st 5 entries are occupied. They are 
/*			a) Null descr
/*			b) kernelds
/*			c) kernelcs
/*      	d) kerneltss
/*			e) threadtss
/*--------------------------------------------*/
        ldtpool_alloc();
        return 0;
}

/*****************************************************
/*	Function		:	gdt_selector_alloc
/*
/*	Callers			:	Any subsystem which wants
/*						allocate a GDT entry. Mostly
/*						called by the arch dependent
/*						process management code.
/*
/*	Functionality	:	Scan thru the "gdtbits" array
/*						searching for a free entry and
/*						return the (free index * 8). Thus
/*						the calling function can directly
/*						OR the RPL and use the result as
/*						a segment selector.
/*
/*	Revision History :	
/*						Sarnath		Feb-14-2002 	Created
/*
/******************************************************/
ulong
gdt_selector_alloc()
{
        int i;
        ulong bit, retval = 0;

        i = 0;
        while ( (i < 256) && ( gdtbits[i] == 0xffffffff ) ) i++;
        if  ( i == 256 )
                return NULL;

        retval = i << 5;
		/*
		 * 1 ULONG has 32-bits and hence
		 * we skip "i * 32" numbers of 
		 * descriptors.
		 */
        bit = 1;
        while ( (gdtbits[i] & bit) ) {
                retval+=1;
                bit = bit << 1;
        }
        if ( bit == 0 ) {
                printk ("\n gdt_selector_alloc: This code should not be executed");
                panic();
        }
        gdtbits[i] |= bit;
        return (retval << 3);
}

/***************************************************************
/*	Function		:	gdt_selector_free
/*
/*	Functionality	:	Unset the bit in the "gdtbits" array
/*						correspoding to the selector.
/*	Revision History:
/*						Sarnath 14-Feb-2002	Created
/*
/*************************************************************/

ulong
gdt_selector_free(ulong selector)
{
        int i;
        ulong bit, byte;

        ASSERT ( (selector >> 3) <= 8192 )
        selector = selector >> 3;
        byte = selector >> 5;
        bit = selector % 32;
        gdtbits[byte] &= ~(1 << bit);   // Set the corresponding bit.
        return 0;
}

/*****************************************************************
 *	Function		:  ldtpool_alloc
 *
 *	Functionality	:  Statically allocate LDTs at boot time. Each process 
 *					   requires only 2 LDT entries. One is for data and other 
 *					   is for code. "phoenix" does not export a segmented model
 *					   to applications. "phoenix" follows a flat model.
 *
 *	Revision History :
 *						Sarnath	Feb-14-2002	Created
 *
 ****************************************************************/

int
ldtpool_alloc()
{
        ldt_pool = (ldt_t *) kmem_align_alloc ( NR_THREADS* sizeof (ldt_t), PAGE);
        return 0;
}


/***********************************************************
/*
/*	Function		:	create_descriptor
/*
/*	Functionality	:	Creates a segment descriptor from the
/*						arguments passed to it.
/*	Revision History :
/*						Sarnath	14-Feb-2002	Created.
/*
/**********************************************************/

int
create_descriptor ( descr, base, limit, system, dpl, type, operandsize, gran )
seg_descr_t *descr;
ulong base,limit,system;
ulong dpl,type,operandsize;
ulong gran;
{
        ASSERT ( descr != NULL )
        descr->limit_0_15 = limit & 0xffff;
        descr->limit_16_19 = ( (limit & 0xf0000) >> 16 );
        descr->base_0_15 = base & 0xffff;
        descr->base_16_23 = ( ( base & 0x00ff0000 ) >> 16 );
        descr->base_24_31 = ( ( base & 0xff000000 ) >> 24 );
        descr->type = type;
        descr->dpl = dpl;
        descr->operand_size = operandsize;
        descr->granualarity = gran;
        descr->system = system;
        descr->zero = 0;
        descr->present = 1;
        return 0;
}

/**************************************************
/*
/*	Function		:	dumpSegmentDescriptor
/*
/*	Functionality	:	Dumps a segment descriptor.
/*						Used for debugging during
/*						early stages.
/*	Revision History :
/*						Sarnath	14-Feb-2002	Created.
/*
/**************************************************/

int
dumpSegmentDescriptor( seg_descr_t *p )
{
        char *ptr;

        ptr = (char *)p;
        printk("\n      SEGMENT DUMP");
        printk("\n Lower Word (limit) = %x", (ulong)(*(unsigned short *)ptr));
        printk("\n Next Word (base) = %x", (ulong) (*(unsigned short *)(ptr + 2)));
        printk("\n Next Byte (base ) = %x", (ulong)(*(ptr + 4)));
        printk("\n Next Byte (type) = %x", (ulong)(*(ptr + 5)));
        printk("\n Next Byte (gran) = %x", (ulong)(*(ptr + 6)));
        printk("\n Next Byte (base ) = %x", (ulong)(*(ptr + 7)));
        printk("\n------------------------------");
        return 0;
}

int dumpTss( struct tss *tssp)
{
	printk("\n Dump of TSS pointer = 0x%x", tssp);	
	printk("\n CR3 = 0x%x", tssp->cr3);
	printk("\n cs = 0x%x", tssp->cs);
	printk("\n ds = 0x%x", tssp->ds);
	printk("\n eip = 0x%x", tssp->eip);
	printk("\n eflags = 0x%x", tssp->eflags);
	return 0;
}
