/*
 *  NOTE: SET TABSTOP TO 4 FOR PROPER VIEW.
 *	
 *	Copyright and Disclaimer								Sarnath, 2002
 *
 *	The software is provided "AS IS".  Absolutely no gaurantee. Use it at
 *	your own risk. The software has been written only for learning purpose.
 *	Thanks for your patience.
 *
 *	ABOUT:
 *
 * 	An O(1) FIRST-FIT kernel memory allocator/de-allocator/coalescer.
 *	Coalescing of free-fragments is done in an O(1) fashion using boundary-tag
 *	idea. (verisimilar to that).
 * 	The idea was construed by me when I felt that the binary buddy algorithm was
 *	wasting a lot of space without using that space to minimize time.
 *	Although it is an original idea (from my perspective), this had already
 *	been construed and implemented in the USER-LAND. I guess LIBC uses a
 *	similar allocator to allocate in USER-SPACE. Another guy called Doug.Lea
 *	has also envisaged a similar idea. You can visit Doug.Lea's page on memory
 *	allocators. He uses a very similar algorithm although his allocator uses a
 *	BEST-FIT strategy. Visit Doug Lea's article on memory allocations
 *	at "http://gee.cs.oswego.edu/dl/html/malloc.html". Doug. Lea's home page
 *	address is "http://g.oswego.edu/dl/".
 *	
 *	REVISION HISTORY:
 *	24-Jul-2002		Sarnath		Created, Coded, Tested the basic version.
 *								Alignment requests are not handled.
 *								Virtual Memory caching is not done explicitly.
 *								However all FREED elements are always added
 *								to the head of the LINKED-LIST. This will
 *								ensure that only the currently freed chunk
 *								will be allocated next.(This may not be cent
 *								percent true if a coalescing happens for a freed
 *								aligned allocation.But thats rare. So we need
 *								not really b bothered about it.)
 *								PHYSICAL memory has not been returned back
 *								on a KMEM_FREE. Some logic has to go in.
 *	27-Jul-2002		Sarnath		Support for ALIGNED requests added. This code
 *								has not been heavily tested.
 *	10-Aug-2002		Sarnath		Bug fixed the small-aligned allocation problem.
 *								Involved just re-arranging the code.
 *  25-Aug-2002		Sarnath		Mapped kernel pages immediately after allocn.
 *								Coded the function "map_kmem_pages" for this.
 *								We wont return failure if we are unable to map
 *								the entire or partial KMEM region. Probably we
 *								must call swapper at that point. This has been
 *								stressed in both "kmem_alloc" and 
 *								"kmem_align_alloc" functions. This will be
 *								incorporated at a later point of time. Changed
 *								the page-fault handler appropriately. See the
 *								comments in the following functions. kmem_alloc,
 *								kmem_align_alloc, kmem_pagefault_handler and
 *								map_kmem_pages.
 *
 * PENDING:
 *
 *	a) The allocator should fail OR sleep (as per an OPTION passed) by the 
 *	   kernel in case physical memory is not available.
 *	b) Unregistering the mapping is not yet done.
 *
 * ENHANCEMENTS TO BE CONSIDERED:
 * 	a)	Prefetching physical memory buffers, Pre-Mapping of VA needs to be 
 *		considered. Thus we need not "pmem_alloc" everytime. (Ofcourse we dont 
 *		do pmem_alloc "everytime" as we do VM caching implicitly.)
 */

/**** Macros and Data Structures required for the main memory allocator ***/
/* { */

#include <generic/types.h>
#include <mm/kmem.h>
#include <mm/vm.h>
#include <mm/pmem.h>

extern ulong freemem;

#ifdef MALLOC_DEBUG
#undef MALLOC_DEBUG
#endif

#define kmalloc			kmem_alloc
#define kfree			kmem_free
#define kmalloc_init	kmem_init
#define kalign_alloc	kmem_align_alloc

#define MALLOC_SIGNATURE	0x0DEADBED
#define SUCCESS				0
#define	FAILURE				!SUCCESS

#define RESOURCE(x)	(((unsigned long)resource) + (unsigned long)(x))
#define OFFSET(x)	((unsigned long)(x) - (unsigned long)(resource))

#define GLOBAL_LIST_UPDATE(ptr, which, what)\
		if (*((unsigned long *)ptr) == MALLOC_SIGNATURE)\
		{\
			struct free_frag *f_macro;\
\
			f_macro = (struct free_frag *)ptr;\
			f_macro->which = what; \
		} else {\
			struct o1suffix *sunique_macro;\
\
			sunique_macro = (struct o1suffix *)ptr;\
			sunique_macro->which = what;	\
		}

#define FREE(x) ((*((unsigned long *)x) == MALLOC_SIGNATURE) ? 1 : 0)

#ifdef MALLOC_DEBUG

#define DUMP_FREE_FRAG(fr)\
{\
	struct free_frag *f_macro;\
\
	f_macro=(struct free_frag *)fr;\
	printk("\n DUMP of free fragment 0x%x", f_macro);\
	printk("\nSignature = 0x%x", f_macro->signature);\
	printk("\nStart = 0x%x", f_macro->start);\
	printk("\nsize = 0x%x", f_macro->size);\
	printk("\nprev = 0x%x", f_macro->prev);\
	printk("\nnext = 0x%x", f_macro->next);\
	printk("\nnext_free = 0x%x", f_macro->next_free);\
	printk("\nprev_free = 0x%x", f_macro->prev_free);\
	printk("\nindex = 0x%x", f_macro->index);\
}

#define DUMP_O1SUFFIX(f)	\
{\
	struct o1suffix *sunique_macro;\
\
	sunique_macro=(struct o1suffix *)f;\
	printk("\n Dump of O1Suffix 0x%x", sunique_macro);\
	printk("\n Next = 0x%x", sunique_macro->next);\
	printk("\n prev = 0x%x", sunique_macro->prev );\
}

#define DUMP_FRAG(f)\
	if (*(unsigned long *)f  == MALLOC_SIGNATURE)\
	{\
		DUMP_FREE_FRAG(f)\
	} else {\
		DUMP_O1SUFFIX(f)\
	}

#else

#define DUMP_FRAG(f) 		;
#define DUMP_O1SUFFIX(x) 	;
#define DUMP_FREE_FRAG(x) 	;

#endif

struct free_frag
{
	unsigned long signature;	/* This has to be the first field. */
	unsigned long start, size;
	void *prev, *next;
	struct free_frag *next_free, *prev_free; 
	int index; /* "index" of the bucket */
};

/*
 *	Data Structure: struct free_frag
 ***********************************
 *
 *	prev,next:
 *	Linked list of all fragments - in ascending order of addresses. 
 *	IRRESPECTIVE of whether they are free or allocated. The pointers are of 
 *	type "free_frag" if the pointer points to private memory. Otherwise it is 
 *	of type "o1suffix". Ordering is maintained by the very nature of memory 
 *	allocation. No explicit sorting is done. It is this aspect that makes 
 *	coalescing an ORDER 1 operation.
 * 
 *	next_free, prev_free:
 *	Linked list of free fragments for this size. Unordered list.
 *
 */

struct free_frag_list
{
	unsigned long size;
	struct free_frag *free_list_head;
	struct free_frag *free_list_tail;
};

/*
 *	Data Structure: free_frag_list (bucket entry)
 ********************************
 *	This data structure forms the entry in each bucket. It contains the linked 
 *	list of free-fragments (in no specific order) for that particular size.
 */

struct o1suffix {
	void *next, *prev;
};

/*
 *	Data Structure: o1suffix
 ***************************
 *	This structure is appened below each allocated memory block. It helps to
 *	maintain a linked list of all fragments in the order of their addresses.
 *	It is this suffix that gives an O(1) coalescing power to this algorithm.
 */

#define NUM_BUCKETS		26
#define WASTE_BUCKET	(NUM_BUCKETS+1)
#define MIN_SIZE		16
#define LOG_MIN_SIZE	4
#define THRESHOLD_SIZE	((unsigned long)(1 << ((NUM_BUCKETS-1)+LOG_MIN_SIZE)))

struct free_frag_list buckets[NUM_BUCKETS + 1];
/*
 *	The extra bucket is for the waste bucket. This bucket holds free fragments
 *  whose sizes are less than MIN_SIZE. They are not considered for any
 *	allocation. They reside in this bucket until they are merged with their
 *	neighbours.
 */

/*** End Data Structures required for the main memory allocator *********/
/* } */

void kmalloc_init();
static void *allocate_private_memory();
static void free_private_memory(void *freed);
static int find_bucket_index(unsigned long size);
static unsigned long break_fragment_at_offset(struct free_frag *frag, 
									unsigned long offset, unsigned long size);

/******* DATA STRUCTURES AND ROUTINES FOR INITIALIZATION AND HANDLING 
         PRIVATE-MEMORY ************/
/* { */

#define MAX_FREE_FRAGMENTS 512
struct private_mem_internals {
	struct free_frag data;
	struct private_mem_internals *next_free;
};
struct private_mem_internals *private_memory, *private_free_head;

#define RESOURCE_SIZE 	 KMEM_SIZE	 /* 256 MB for kernel memory */
void *resource;

int kmalloc_init_happened = 0;

void
kmalloc_init()
{
	int i,index;
	struct free_frag *temp;

	for(i=0; i< NUM_BUCKETS; i++)
	{
		buckets[i].size = (1 << (i+LOG_MIN_SIZE));
		buckets[i].free_list_head = NULL;
		buckets[i].free_list_tail = NULL;
	}

	buckets[WASTE_BUCKET-1].size = 0;
	buckets[WASTE_BUCKET-1].free_list_head = NULL;
	buckets[WASTE_BUCKET-1].free_list_tail = NULL;
/*
 *	Scope: The code above
 *	The waste bucket holds free-fragments which are less than MIN_SIZE. They 
 *	will not be considered for allocatins. These fragments sit there until they
 *	get coalesced with their neighbours.
 */

/*----------Allocate Internal Memory by deliberately page faulting --------*/

	statickernelseg.type = MAP;
	statickernelseg.pageFaultHandler = &map_segment_handler;
	private_memory = (struct private_mem_internals *) (statickernelseg.base + 
													statickernelseg.limit);
	statickernelseg.limit += MAX_FREE_FRAGMENTS * 
							 sizeof(struct private_mem_internals);
	
	makeZero((char *)private_memory,
			MAX_FREE_FRAGMENTS*sizeof(struct private_mem_internals));

	statickernelseg.type = PANIC;
	statickernelseg.pageFaultHandler = &panic_segment_handler;

/*----------End Allocate Internal Memory ---------------------------------*/

	for(i=0; i<MAX_FREE_FRAGMENTS-1; i++)
	{
		private_memory[i].next_free = &private_memory[i+1];
		private_memory[i].data.signature = MALLOC_SIGNATURE;
	}
	private_memory[i].next_free = NULL;
	private_free_head = &(private_memory[0]);

	resource = (void *)KMEM_BASE_VA;  

	/* Allocate the Master free fragment - The parent of the Big Bang. */

	temp = (struct free_frag *) allocate_private_memory();
	index= find_bucket_index(RESOURCE_SIZE);
	buckets[index].free_list_head = temp;
	buckets[index].free_list_tail = temp;
	if (temp == NULL)
	{
		panic("\n First allocation to private memory itself failed");
	}
	
	temp->start = 0;
	temp->size =  RESOURCE_SIZE;
	temp->prev = NULL;
	temp->next = NULL;
	temp->next_free = NULL;
	temp->prev_free = NULL;
	temp->index = index;

	kmalloc_init_happened=1;
	return;
}

static void *
allocate_private_memory()
{
	void *temp;

	if (private_free_head)
 	{
		temp = (void *)private_free_head;
		private_free_head = private_free_head->next_free;
		return temp;
	}
	return NULL;
}

static void
free_private_memory(void *freed)
{
	struct private_mem_internals *temp;

	temp = (struct private_mem_internals *)freed;

	if (private_free_head == NULL)
	{
		private_free_head = temp;
		temp->next_free = NULL;
		return;
	}
	
	temp->next_free = private_free_head;
	private_free_head = temp;
	return;
}

/*********** End INIT and Private Mem Handling Routines *********************/
/* } */


/********* Support Functions and the Main memory allocator *****************/
/* { */

/**************************************************************************
 * function: find_bucket_index(unsigned long size)
 *
 * functionality:
 * The function uses binary search to find the appropriate bucket for the 
 * requested size. For example, it returns 128th bucket index for any
 * size between 128 to 255. The time-complexity being: O(log(NUM_BUCKETS)); 
 * Never more than 5 for 32-buckets.
 *
 **************************************************************************/
static int 
find_bucket_index(unsigned long size)
{
	int lo,hi;

	lo=NUM_BUCKETS/2;
	hi=NUM_BUCKETS-1;

	while(1)
	{
		if (size == buckets[lo].size)
			return lo;

		if (size > buckets[lo].size)
		{
			int temp;

			if (lo == (NUM_BUCKETS-1))
				return lo;
			
			temp = lo;
			lo=(lo+hi)/2;
			if ( lo == temp) 
			{
				lo=hi;
			}
			continue;
		}

		if (lo != 0)
		{
			if (size > buckets[lo-1].size) {
				return lo-1;
			}

			hi=lo;
			lo=lo/2;
		}
		else
		{
			printk("\n We should never get his condition. The given size");
			printk(" should be greater than MIN_SIZE");
			printk("\n The given size = 0x%x", size);
			return lo;
		}
	}
}

/*****************************************************************************
 * Function: static struct free_frag * find_frag(struct free_frag *p, 
 *												 unsigned long size)
 *
 * Functionality: 
 * This function searches a linked list (un-ordered) for a fragment of size
 * >= requested size. Time complexity = O(length of list). However this search 
 * can be avoided by increasing NUM_BUCKETS so that we handle sizes >= 
 * RESOURCE_SIZE in an O(1) way without calling this function. If we use lesser 
 * NUM_BUCKETS then this code will be hit for larger sizes. The NUM_BUCKETS is 
 * a tuning parameter for this allocator. For a 4GB RESOURCE_SIZE and 
 * NUM_BUCKETS of 28 this code will never need to be called. So dont jump into 
 * conclusion that the allocator is in-efficient after seeing the time
 * complexity associated with this allocator.
 *****************************************************************************/ 
static struct free_frag *
find_frag(struct free_frag *p, unsigned long size)
{
	while (p)
	{
		if (p->size >= size)
			return p;
		p = p->next_free;
	}
	return NULL;
}

/******************************************************************************
 * Function:	static void detach_from_free_list(struct free_frag *frag);
 * 
 * Functionality:
 * As expected, this function just deletes a free fragment from the list.
 * Involves basic doubly-linked list deletion code. 
 *****************************************************************************/
static void
detach_from_free_list(struct free_frag *frag)
{
	struct free_frag_list *f;

	f = &buckets[frag->index];

	if (f->free_list_head == frag)
		f->free_list_head = frag->next_free;
	if (f->free_list_tail == frag)
		f->free_list_tail = frag->prev_free;

	if (frag->prev_free)
	{
		frag->prev_free->next_free = frag->next_free;
	}

	if(frag->next_free)
	{
		frag->next_free->prev_free = frag->prev_free;
	}

	frag->next_free = NULL;
	frag->prev_free = NULL;
	return;
}

/*****************************************************************************
 * Function: enqueue_in_free_list(struct free_frag *frag, int index)
 *
 * Functionality:
 * As expected, this function moves a fragment from one list to another.
 * The caller specifies the "index" of the list to which the fragment should
 * be enqueued. The code just adds the fragment to the tail of the list.
 * Involves mere doubly-linked list addition.
 *
 *****************************************************************************/
static void
enqueue_in_free_list(struct free_frag *frag, int index)
{
	struct free_frag_list *f = &buckets[index];

	frag->index = index;

	if ((f->free_list_tail == NULL) || (f->free_list_head == NULL))
	{
		f->free_list_tail = frag;
		f->free_list_head = frag;
		frag->next_free = NULL;
		frag->prev_free = NULL;
		return;
	}

	ASSERT(f->free_list_tail->next_free == NULL);
	f->free_list_tail->next_free= frag;
	frag->prev_free = f->free_list_tail;
	frag->next_free = NULL;
	f->free_list_tail = frag;
	return;
}

/*****************************************************************************
 * Function: attach_to_free_list(struct free_frag *frag, unsigned long index)
 *
 * Functionality:
 * The functionality is similar to the "enqueue" function. However this
 * function adds the fragment to the head of the list rather than the tail.
 *
 **************************************************************************/
static void
attach_to_free_list(struct free_frag *frag, unsigned long index)
{
	struct free_frag_list *f;

	f=&buckets[index];
	frag->index=index;
	if (f->free_list_head == NULL)
	{
		f->free_list_head = frag;
		f->free_list_tail = frag;
		frag->next_free = NULL;
		frag->prev_free = NULL;
		return;
	}

	frag->next_free = f->free_list_head;
	frag->prev_free = NULL;
	f->free_list_head->prev=frag;
	f->free_list_head = frag;
	return;
}

/*****************************************************************************
 * Function: break_fragment( struct free_frag_list *, unsigned long size,
 *							 unsigned long align )
 *
 * Caller: "kmalloc" calls this function to allocate memory.
 *
 * Functionality:
 * This function does the core of the functionality of "kmalloc". The function
 * breaks "size" bytes from the fragment "free_frag_list" and returns the start
 * address of the fragment if the alignment specified is zero. However
 * If "align" is specified, the routine returns an aligned value. The "align"
 * must be a power of 2.
 *
 * Assumption:
 * The caller has to make sure that the "free_frag_list *" contains fragments
 * which are greater than size "size".
 *
 *****************************************************************************/
static unsigned long
break_fragment( struct free_frag_list *p, unsigned long size, 
				unsigned long align)
{
	struct free_frag *frag;
	int index;
	unsigned long retval=0, offset;

	if ((align & (align-1)) != 0)
	{
		printk("\nbreak_fragment: ALIGN Argument is not a power of 2");
		return NULL;
	}

/*
 * First find out the fragment that can support it.
 */
	if (p->size == THRESHOLD_SIZE)
	{
		if (align != 0)
		{
			return NULL;
		}

		printk("\n This code will never be reached");
		printk(" Because THRESHOLD_SIZE is > RESOURCE_SIZE\n");

		frag = find_frag(p->free_list_head,size);
		if (frag == NULL)
		{
			return NULL;
		}
	} else {
		frag = p->free_list_head;
	}

	if (frag == NULL) 
	{
/* There is still a chance of being NULL. Checkout the caller */
		return NULL;
	}

	ASSERT(frag->size >= (size+align));

/*
 * According to the alignment specified, find out where to break the fragment.
 */
	retval = frag->start;
	offset = 0;
	if (align)
	{
		offset = retval & (align-1);
		if (offset)
		{
			offset = align - offset;
		}
	}
	retval = break_fragment_at_offset(frag, offset, size);
	return retval;
}

static unsigned long
break_fragment_at_offset(struct free_frag *frag, unsigned long offset,
						unsigned long size)
{
	unsigned long retval, orig_size_of_free_fragment;
	struct free_frag *f1=frag, *f3, *f1_original_next;
	struct o1suffix *suffix;
	int index;

	retval = f1->start + offset;

	if (offset == 0)
	{
		frag->size 	-= size;
		frag->start += size;
	
		if (frag->size != 0)
		{
			/*
			 *	Now that the fragment has shrinked, move it to an approriate
			 *	position if necessary.
			 */
			if (frag->size < MIN_SIZE)
			{
				index = WASTE_BUCKET;
			} else {
				index = find_bucket_index(frag->size);
			}

			if (index != frag->index)
			{
				detach_from_free_list(frag);
				enqueue_in_free_list(frag, index);
			}
	
			/*
			 *	Update the Global List of fragments. The updation maintains the
			 *	order of addresses.
			 */
			suffix = (struct o1suffix *)( RESOURCE(retval) + size - 
										  sizeof(struct o1suffix) );
			suffix->next = frag;
			suffix->prev = frag->prev;
			if (frag->prev)
			{
				GLOBAL_LIST_UPDATE(frag->prev, next, suffix);
			}
			frag->prev = suffix;
		} else {
			/*
			 * The fragment has been completely exhausted. Update the Global
			 * list of fragments. This updation maintains the order of 
			 * addresses.
			 */
			suffix = (struct o1suffix *)( RESOURCE(retval) + size - 
										  sizeof(struct o1suffix));
			suffix->next = frag->next;
			suffix->prev = frag->prev;
			/*
			 * CRITICAL: The above pointer population will cause page-faults
			 * if there is no mapping. If there are no physical pages available
			 * the page-fault handler will not provide the required mapping.
			 * In such a case, this piece of code will cause recursive page
			 * faults. This needs to be handled specially. We might consider
			 * sleeping for physical pages in such a case and call the
			 * swapper.
			 */
			if (frag->prev)
			{
				GLOBAL_LIST_UPDATE(frag->prev, next, suffix);
			}
			if (frag->next)
			{
				GLOBAL_LIST_UPDATE(frag->next, prev, suffix);
			}
			frag->next = NULL;
			frag->prev = NULL;
			free_private_memory(frag);
		}
	} else {
	/*
 	 * In this case, the fragment is broken in the middle. We may get 2 free
	 * fragments + 1 allocated fragment OR We may get a free fragment + 1
	 * allocated fragment. We need to handle both the cases.
	 */
		#ifdef MALIGN_DEBUG
			printk("\nretval = 0x%x", retval);
		#endif
		orig_size_of_free_fragment = frag->size;
		f1 = frag;
		f1->size = offset;
		f1_original_next = f1->next;
		suffix = (struct o1suffix *) (RESOURCE(retval) + size - 
									  sizeof(struct o1suffix));
		#ifdef MALIGN_DEBUG
			printk("\n Suffix = 0x%x\n", suffix);
		#endif
		suffix->next = f1->next;
		suffix->prev = f1;
		f1->next = suffix;

	/*
	 * Now since "f1" size has changed, attach him to the correct bucket.
	 */
		if (f1->size < MIN_SIZE)
        {
        	index = WASTE_BUCKET;
			/*
			 * An appropriate "colaescing" operation  will salvage this guy 
			 * from the WASTE_BUCKET.
			 */
        } else {
        	index = find_bucket_index(f1->size);
        }

        if (index != f1->index)
        {
            detach_from_free_list(f1);
        	enqueue_in_free_list(f1, index);
        }

		ASSERT( (offset + size) <= orig_size_of_free_fragment )

		if ((offset + size) < orig_size_of_free_fragment)
		{
			#ifdef MALIGN_DEBUG
				printk("\n offset + size < (orig_size=0x%x)", 
						   orig_size_of_free_fragment);
			#endif
			f3 = allocate_private_memory();
			if (f3 == NULL)
			{
				printk("\nbreak_fragment_at_offset: Unable to allocate ");
				printk("private memory");
				return NULL;
			}
			#ifdef PARANOID
				f3->signature = MALLOC_SIGNATURE;
			#endif
			f3->start = f1->start+offset+size;
			f3->size = orig_size_of_free_fragment - f3->start;
			f3->prev = suffix;
			f3->next = f1_original_next;
			suffix->next = f3;
			if (f3->next)
			{
				GLOBAL_LIST_UPDATE(f3->next, prev, f3)
			}
		    #ifdef MALIGN_DEBUG 
			   else {
				printk("\n f3->next == NULL");
			}
			#endif
			
			/*
			 * Now that we have added F3 to the Global List, we need
			 * to add to the appropriate bucket.
			 */
			index = find_bucket_index(f3->size);
			enqueue_in_free_list(f3, index);

			#ifdef MALIGN_DEBUG
				printk("\n Summary: ");	
				printk("\n f1 = 0x%x", (unsigned long)(f1));
				printk("\n f1->next = 0x%x", (unsigned long)(f1->next));
				printk("\n f1->prev = 0x%x", (unsigned long)(f1->prev));
				printk("\n f1->size = 0x%x", (unsigned long)(f1->size));
				printk("\n f1->start = 0x%x", (unsigned long)(f1->start));
				printk("\n suffix = 0x%x", (unsigned long)(suffix));
				printk("\n suffix->next = 0x%x", (unsigned long)(suffix->next));
				printk("\n suffix->prev = 0x%x", (unsigned long)(suffix->prev));
				printk("\n f3 = 0x%x", (unsigned long)(f3));
				printk("\n f3->next = 0x%x", (unsigned long)(f3->next));
				printk("\n f3->prev = 0x%x", (unsigned long)(f3->prev));
				printk("\n f3->size = 0x%x", (unsigned long)(f3->size));
				printk("\n f3->start = 0x%x", (unsigned long)(f3->start));
			#endif
		} else {
			/*
			 * We dont have any problems here. We just leave it like that.
			 * Everything is in a SANE condition. The comment is here only
			 * for clarity (for source browsers)
			 */
			  #ifdef MALIGN_DEBUG
                printk("\n offset + size == (orig_size=0x%x)",
                           orig_size_of_free_fragment);
              #endif
		}
	}
	return	RESOURCE(retval);
}

/******************************************************************************
 * Function: kmalloc(unsigned long size)
 *
 * Functionality:
 * This is the interface function provided to the rest of the kernel to
 * allocate kernel memory.
 *****************************************************************************/
unsigned long
kmalloc(unsigned long size)
{
	int index, original_index;
	unsigned long retval ;
	int map_retval;

	ASSERT(kmalloc_init_happened != 0)

	if (size > (freemem*PAGE))
	{
		printk("\nUnable to service the request due to less physical memory");
		/*
		 * FUTURE TODO (VERY IMPERATIVE): 25-Aug-2002
		 * We should decide either to sleep or fail according to some kmem flag
		 * passed. We should also invoke the swapper at this point if the swap
 		 * subsystem in place. 
		 */ 
		return NULL;
	}

	/*
	 * FUTURE TODO: 25-Aug-2002
	 * We should obtain an allocation lock over the physical memory subsystem 
	 * until the end because some one else (may b an interrupt handler or 
 	 * another thread running in another processor) may swallow memory before 
	 * we could really consume it. All these have been deferred until the 
	 * kernel grows to that level.
	 */

	if (size < MIN_SIZE)
	{
		size = MIN_SIZE;
	}
	size += sizeof(struct o1suffix);

	if (size > RESOURCE_SIZE)
	{
		printk("\n Unable to handle BIG size");
		return NULL;
	}

	index = find_bucket_index(size);
	original_index = index;

	while( (buckets[index].free_list_head == NULL) && (index < NUM_BUCKETS))
	{
		index++;
	}

	if (index == NUM_BUCKETS)
	{
	#ifdef MALLOC_DEBUG
		printk("\nkmalloc: No free memory available.");
	#endif
		return NULL;
	}

	if ((original_index == index) && ((index+1) != NUM_BUCKETS))
	{
		index++;
	}

	retval = break_fragment(&buckets[index], size,0);
	if (retval != NULL)
	{
		map_retval = map_kmem_pages(retval, size);
		if ( map_retval == -1)
		{
			/*
			 * FUTURE TODO: 25-Aug-2002
			 * Either the ENTIRE or PARTIAL set of the allocated memory was
			 * unable to be mapped. We need to call swapper or do sthg else 
 			 * here.
			 * Note that this condition is RARE but CAN HAPPEN if we do not
			 * hold the PHYS_MEM_ALLOCATION_LOCK. See the comment at the
			 * begining of the function.
			 */
			printk("\nnmap_kmem_pages failed\n");
		}
	}
	return retval;
}


/****************************************************************************
 * Function: kalign_alloc(unsigned long size, unsigned long align)
 *
 * Functionality:
 * This function aids aligned allocation.
 ****************************************************************************/
unsigned long
kalign_alloc(unsigned long size, unsigned long align)
{
	unsigned long  orig_size = size;
	int orig_index, index, map_retval;
	ulong retval;

	if (align == 0)
	{
		printk("\nkmem_align_alloc: The \'align\' argument is zero.");
		return NULL;
	}

	if ( (align & (align -1)) !=0 )
	{
		printk("\nkmem_align_alloc: The \'align\' argument should be a");
		printk(" power of two.");
		return NULL;
	}

	if (size > freemem)
    {
        printk("\nUnable to service the request due to less physical memory");
        /*
         * FUTURE TODO: 25-Aug-2002
         * We should decide either to sleep or fail according to some kmem flag
         * passed. We should also invoke the swapper at this point if the swap
         * subsystem in place.
		 */
        return NULL;
    }

   /*
 	* FUTURE TODO: 25-Aug-2002
	* We should obtain an allocation lock over the physical memory subsystem 
	* until the end because some one else (may be an interrupt handler or 
    * another thread running in another processor) may swallow physical memory 
    * before we could really consume it. All these have been deferred until the
    * kernel grows to that level.
    */

	if (size < MIN_SIZE)
	{
		size = MIN_SIZE;
		orig_size = size;
	}
	size += align;
	size += sizeof(struct o1suffix);
	if (size > RESOURCE_SIZE)
	{
		printk("\nkmem_align_alloc: Unable to handle this big size");
		return NULL;
	}
	
	orig_index = find_bucket_index(size);
	index = orig_index;

	while( (buckets[index].free_list_head == NULL) && (index < NUM_BUCKETS))
    {
        index++;
    }

    if (index == NUM_BUCKETS)
    {
    #ifdef MALLOC_DEBUG
        printk("\nkmalloc: No free memory available.");
    #endif
        return NULL;
    }

    if ((orig_index == index) && ((index+1) != NUM_BUCKETS))
    {
        index++;
    }

    retval = break_fragment(&buckets[index], (orig_size + sizeof(struct o1suffix)), align);
	if (retval != NULL)
	{
		map_retval = map_kmem_pages(retval,orig_size);
		if ( map_retval == -1)
        {
            /*
             * FUTURE TODO: 25-Aug-2002
             * Either the ENTIRE or PARTIAL set of the allocated memory was
             * unable to be mapped. We need to call swapper or do sthg here.
             * Note that this condition is RARE but CAN HAPPEN if we do not
             * hold the PHYS_MEM_ALLOCATION_LOCK. See the comment at the
             * begining of the function.
             */
            printk("\nnmap_kmem_pages failed\n");
        }
	}
    return retval;
}


/****************************************************************************
 * Function: int map_kmem_pages(unsigned long retval, unsigned long size)
 *
 * Functionality:
 * The function maps the kernel-memory that has been allocated via "kmem_alloc"
 * or "kmem_align_alloc". This is done in order to avoid page-faults
 * (additional overhead). Previously, we did not have such an approach. The
 * kernel memory was mapped in a delayed way previously. But as pointed out by
 * Deepak veliath, delayed mappings incur only overhead if we are sure that the
 * page-fault is going to happen. Delayed mapping  is applicable only to unsure
 * entities like "an user mmaped file", "User code", "User data" etc.
 *
 * Revision History:
 * 25-Aug-2002	Sarnath		Coded the first version.
 *
 ***************************************************************************/
int
map_kmem_pages(ulong retval, ulong size)
{
	ulong covered = 0, start,end, mapping, pa;

	start = PAGE_ALIGN(retval);
	end = PAGE_ALIGN((retval+size));
	do
	{
		mapping = get_pa(start, &kas);
		if (mapping == 0xffffffff)
		{
			pa = pmem_alloc(SYSTEM);
			if (pa == PA_MAX)
			{
				return -1;
			}
			map(start, pa, &kas);
		}
		start += PAGE;
	} while(start <= end);
	return 0;
}

/****************************************************************************
 * Function: coalescable(void *ptr, unsigned long size) 
 *
 * Functionality:
 * The function returns SUCCESS if the freed kernel memory (ptr) can be 
 * coalesced with any of its neighbouring chunk. Otherwise it returns FAILURE.
 ***************************************************************************/
static inline int
coalescable(void *ptr, unsigned long size)
{
	struct o1suffix *s;

	s=(struct o1suffix *)((char *)ptr+size);
	if (((s->next != NULL) && FREE(s->next)) || ((s->prev != NULL) && FREE(s->prev)))
		return SUCCESS;
	return FAILURE;
}

/******************************************************************************
 * Function: static void coalesce(void *ptr, unsigned long size)
 *
 * Functionality:
 * The argument passed is the freed memory pointer and freed size. There can
 * be a maximum of only 2 coalescing operations possible. This is because:
 *	   a) At any point of time the global-list is a condensed list. i.e no
 *	   	  coalecing is possible on the list unless a FREE happens.
 *	   b) The introduction of a new Free Chunk F2 between F1 and F3 can cause
 *		  a coalescing of F1-F2-F3 into 1 chunk. If F3 or F1 is not a free
 *		  fragment then we can have only 1 coalesce operation. i.e F1-F2 or
 *		  F2-F3. If both F1 and F3 are allocated chunks then we dont need to
 *		  coalesce anything.
 * Notion:
 * The following notion is followed in the comments:
 *	F2 ->  Refers to the fragment being freed.
 *	F1 ->  F2's Previous Fragment.
 *  F3 ->  F2's Next Fragment.
 *****************************************************************************/
static void
coalesce(void *ptr, unsigned long size)
{
	struct o1suffix *s;
	struct free_frag *f1, *f3, *affected;
	char f1_coalesced = 0, f3_coalesced = 0;
	int index;

	s = (struct o1suffix *)((char *)ptr + size);
	f1 = s->prev;
	f3 = s->next;

	#ifdef MALLOC_DEBUG
	printk("\n f1 = 0x%x, f3=0x%x", f1, f3);
	#endif

	if (f1 != NULL)
	{
		if (FREE(f1))
		{
			#ifdef MALLOC_DEBUG
			printk("\n A colescable F1 found");
			#endif
			/* Coalesce F1 and F2 */
			f1_coalesced = 1;
			f1->size += (size + sizeof(struct o1suffix));
			f1->next = s->next;
			if (s->next)
			{
				GLOBAL_LIST_UPDATE(s->next, prev, f1);
			}
		}
	}

	if (f3 != NULL)
	{
		if (FREE(f3))
		{
			#ifdef MALLOC_DEBUG
			printk("\ncoalesce: Free F3 Fragment found");
			#endif
			if (f1_coalesced)
			{
				/* Now F1 and F3 had to be coalesced. */
				f1->size += (f3->size);
	            f1->next = f3->next;
	            if (f3->next)
	            {
	                GLOBAL_LIST_UPDATE(f3->next, prev, f1);
	            }
				free_private_memory((void *)f3);
			} else {
			   /* Coalesce F2 and F3 */
				f3->size += (size + sizeof(struct o1suffix));
				f3->start = OFFSET(ptr);
				if (s->prev)
				{
					GLOBAL_LIST_UPDATE(s->prev, next, f3);
				}
				f3->prev = s->prev;
				f3_coalesced = 1;
			}
		}
	}

	/*
	 * Now either F1 or F3 or none has changed in size. So fit them in the
	 * correct bucket.
	 */
	if (f1_coalesced)
	{
		affected = f1;
	}
	else if (f3_coalesced)
	{
		affected = f3;
	}
	
	if (affected == NULL)
	{
		return;
	}
	
	index = find_bucket_index(affected->size);
	if (index != affected->index)
	{
		detach_from_free_list(affected);
		#ifdef MALLOC_DEBUG
		printk("\nAffected fragment moved from index %d to index %d", 
				affected->index,index);
		#endif
		attach_to_free_list(affected, index);	
	}
	return;
}

/******************************************************************************
 * Function: kfree(void *ptr, unsigned long size)
 *
 * Functionality:
 * This function is an interface function which is exported to the rest of the
 * kernel. The function also takes care of coalescing the chunk with its 
 * neighbours.
 *****************************************************************************/
int
kfree(void *ptr, unsigned long size)
{
	struct free_frag *host;
	struct o1suffix *s;
	unsigned long index;

	ASSERT(size <= RESOURCE_SIZE)
	if (size < MIN_SIZE)
	{
		size=MIN_SIZE;
	}

	s=(struct o1suffix *)( ((unsigned long)ptr)+size);

	if (coalescable(ptr,size) != SUCCESS)
	{
		#ifdef MALLOC_DEBUG
		printk("\nkfree: No Coalescable fragment found.");
		#endif
		host= (struct free_frag *)allocate_private_memory();
		if (host == NULL)
		{
			printk("\nkfree: Unable to allocate internal memory");
			return -1;
		}

		ASSERT(host->signature == MALLOC_SIGNATURE)
		host->start = OFFSET(ptr);
		host->size = size+sizeof(struct o1suffix);

		/* Update the Global List appropriately */
		if (s->prev)
		{
			GLOBAL_LIST_UPDATE(s->prev,next,host);
		}

		if (s->next)
		{
			GLOBAL_LIST_UPDATE(s->next,prev,host);
		}
		host->next=s->next;
		host->prev=s->prev;

		/* Insert the free fragment into an appropriate list. */
		index=find_bucket_index(host->size);
		attach_to_free_list(host,index);
		return SUCCESS;
	} else {
		#ifdef MALLOC_DEBUG
		printk("\nkfree: coalescable fragment found. Proceeding to coalesce");
		#endif
		coalesce(ptr,size);
	}
	return SUCCESS;
}


/******** END Support Functions and the Main memory allocator ***************/
/* } */

/******************* KMEM SEGMENT PAGE FAULT HANDLER ************************/

int
kmemsegment_pagefault_handler(ulong faulted_va)
{
	ulong pa;
	int	retval;

	printk("\nKMEM_PAGEFAULT_HANDLER INVOKED\n");

	/*  Date: 25-Aug-2002
	 *	We do all the MAPPING during the kernel memory allocation time itself, 
 	 *	synchronously. But there is a possibility that "map_kmem_pages" would 
	 *	have failed during allocation time. Most probably physical mem would
	 *	be very scarce at this point. Anyway we will try for it once before
	 *  finally giving up.
	 *
 	 * 	FUTURE TODO:
	 *  One day this handler will just do a PANIC and nothing else. All these
	 *	code will go away.
	 */
	pa = pmem_alloc(SYSTEM);
	if (pa == 0xFFFFFFFF)
	{
		printk("\nkmemsegment_pagefault_handler: Not enough physical memory");
		/*
		 * FUTURE TODO:
		 * We should call swapper at this point to salvage physical memory
		 * for us.
		 */
		panic();
	}
	retval = map(PAGE_ALIGN(faulted_va), pa, &kas);
	return retval;
}
/******************* END KMEM SEGMENT HANDLER ********************************/
