/*
 *      Author: K.Sarnath. 27-Jan-2001
 *      Kernel Virtual Memory Manager.
 *      Kernel page fault handler is present in this file.
 *      Refer (config.h) for details about how kernel VA is split.
 */

#include <mm/vm.h>
#include <config.h>
#include <generic/types.h>
#include <i386/page.h>
#include <mm/pmem.h>
#include <mm/vm.h>

/*
 *      Describe the kernel Virtual Address Space
 *      See <config.h> for more details
 *      The file "include/mm/vm.h" -> contains defns of "as" and "seg"
 *      structures.
 */
seg_t panicseg;
as_t kas = {&panicseg, (pte_t*) (KERNEL+KPGD) };
seg_t kmemseg = { KMEM_SEG_BASE, KMEM_LIMIT, KMEM , NULL,NULL };
seg_t statickernelseg = { STATIC_KERNEL, STATIC_KERNEL_LIMIT, PANIC, &kmemseg,
                          NULL};
seg_t videosegments =  { KERNEL + START_VIDEOBUFFERS, VIDEOLIMIT, PANIC,
                         &statickernelseg , NULL};
seg_t panicseg = { KERNEL, FREEMEM, PANIC, &videosegments, NULL};

/************** End of describing KVA *******************/

extern pte_t *kpgd;
extern pmem_t *mem_table;
extern int early_stage;
int panic_segment_handler ();
int map_segment_handler ();
void * get_temp_mapping ( ulong pa ) ;


void
kseg_init()
{
        kmemseg.pageFaultHandler = &kmemsegment_pagefault_handler;
        statickernelseg.pageFaultHandler=&panic_segment_handler;
        videosegments.pageFaultHandler=&panic_segment_handler;
        panicseg.pageFaultHandler=&panic_segment_handler;
}

ulong
get_pa ( ulong va , as_t *as)
{
        uint pd_offset, pt_offset;
        ulong ptable ,retval;
        uint nonpaged_address_part;
        pte_t *ptr = (pte_t*) NULL, *pgdir;

        if ( as == NULL ) {
                printk("\n PANIC: get_pa : as = NULL ");
                panic();
        }
        pgdir = as->pgdir;
        nonpaged_address_part = ( va & (PAGE -1) );
        pd_offset = PGDIR_OFFSET (va);
        pt_offset = PGTABLE_OFFSET (va);
        if ( ( pgdir[pd_offset].pte & PG_PRESENT ) == NULL ) {
        #ifdef __DEBUG__
                printk("\n get_pa: Invalid VA passed. Non present page ");
        #endif
                return PA_MAX;
        }
        ptable = PAGE_ALIGN(pgdir[pd_offset].pte);
        ptr = (pte_t *)get_temp_mapping (ptable);
        if ( ( ptr[pt_offset].pte & PG_PRESENT ) == NULL ) {
        #ifdef __DEBUG__
                printk("\n get_pa: Invalid VA passed. Non present page ");
        #endif
                return PA_MAX;
        }
        retval = PAGE_ALIGN (ptr[pt_offset].pte);
        retval |= nonpaged_address_part;
        return retval;
}


void
print_kernel_segments ()
{
        as_t *tr;
        seg_t *temp;

        printk ("\nPrinting Kernel Segments \n");
        tr = &kas;
        temp = tr->segments;

        while ( temp ) {
                printk ("BASE: %x\tLimit: %x\n", temp->base , temp->limit);
                temp = temp->next;
        }
        printk ("\nEnd printing kernel segments\n");
}

/*
 *   Note:
 *        Be careful while using this function. This function Maps a constant
 *   kernal virt address, say free_virt, to the given physical address.
 *   Suppose, u use "get_a_temp_mapping in function x, after getting it
 *  u call function "y" , which also gets a temp_mapping, then the latter
 *  destroys the mapping of the former. So, save n restore mappings in that
 *  case.
 */
void *
get_temp_mapping ( ulong pa )
{
        ulong va;
        pte_t *tva;
        ulong pd_offset, pt_offset, ptable, off ;

        off = pa & ( PAGE - 1 );
        pa = PAGE_ALIGN (pa);
        va = FREE_VIRT;

/*
 *      The purpose of FREE_VIRT is that, u can make this free VIRT to
 *      point to any Physical address. This helps us to populate page
 *      tables and other objects, which do not have a virtual mapping
 *                                                      - Sarnath.K
 */

        pd_offset = PGDIR_OFFSET(va);
        pt_offset = PGTABLE_OFFSET(va);
        ptable = kpgd[pd_offset].pte;
        ptable = PAGE_ALIGN(ptable);
        tva = (pte_t *)(ptable + KERNEL);
/*
 *      I get VA for this PA directly to avoid chicken-egg problem.
 */
        tva = tva + pt_offset;
        tva->pte =  (pa | PG_PRESENT | PG_WRITE);
        flush_tlb();
        va = va + off;
        return (void *)va;
}

/*
 * This guy is the mapping function. ( VtoP)
 * 386 features.
 */
int
map (ulong va, ulong pa , as_t *as ) // Maps only a single page.
{
        void *tva;
        ulong pgd_offset, pt_offset;
        static pte_t *ptable = NULL;    //Note: Its static.
        pte_t *pgd;

        if ( ( as == NULL ) || ( as->pgdir == NULL ) ) {
                printk (" MAP: \"as\" is invalid \n");
                panic();
        }
        pgd = as->pgdir;
        pgd_offset = PGDIR_OFFSET(va);
        if ( pgd[ pgd_offset ].pte == 0 ) {
/*
 * Now we need to allocate a page table
 */
                if ( early_stage ) {
#ifdef __PAGEFAULTDEBUG__
printk ("\n MAP: Allocating a page table..... in early_stage\n");
#endif
                        if ( !ptable ) ptable = (pte_t*) (EARLY_PTABLE);
                        else ptable = (pte_t*) ((char *)ptable + PAGE);
                        if ( (ulong) ptable >= FREEMEM ) {
                                printk ("\n Unable to handle this big memory. \n Need to reconfigure the kernel with more space for early page tables.....\nReboot manually \n");
                                return -1;
                        }
                } else {
                        /*
                         * Allocate a page table. Use pmem_alloc
						 *
						 * FUTURE NOTE:
						 * The Page-Directories of all address-spaces who share
						 * the kernel address-space should be updated if the
						 * page-table is allocated for the kernel address space.
                         */
                        ptable = (pte_t *) pmem_zalloc (PAGE_TABLE);
                        if ( (ulong) ptable == PA_MAX ) {
                                printk ("\n Kernel: Out of physical memory\n");
                                return -1;
                        }
#ifdef __PTOVDEBUG__
        printk ("\n Allocating page table = %x", ptable );
#endif
                }
#ifdef __PAGEFAULTDEBUG__
        printk ("\n MAP: Populating the page directory.\t");
        printk ("MAP: Pgdir address %x Ptable address %x ", (ulong) pgd , (ulong) ptable );
#endif
                pgd [pgd_offset].pte = ( (ulong)ptable | PG_PRESENT  | PG_WRITE );
                if ( va < KERNEL ) {
                        pgd [pgd_offset].pte |= PG_USER;
                }
        }
        else {
                ptable = (pte_t *) PAGE_ALIGN(pgd[pgd_offset].pte);
        }
/*
 * Now we need to populate the page table using a temp mapping
 */
                tva = get_temp_mapping ( (ulong)ptable );
                pt_offset = PGTABLE_OFFSET(va);
#ifdef __PAGEFAULTDEBUG__
        printk (" pa = %x , tva = %x, pagetable = %x, pt_offset = %x\n", pa , (ulong)tva , ptable, pt_offset);
#endif
                pa = PAGE_ALIGN(pa);
                tva = (void *) ((pte_t *)tva + pt_offset );
                ((pte_t*)tva)->pte = (  pa | PG_PRESENT | PG_WRITE );
                if ( va < KERNEL ) {
                        ((pte_t*)tva)->pte |= PG_USER;
                }
#ifdef __PAGEFAULTDEBUG__
        printk ("MAP: Everyting is over. :) \n");
#endif
                flush_tlb();
                register_mapping(va, pa, as);
                return 0;
}

/*
 * This guy unloads mapping for a virtual address, in an address space.
 * Machine depedendant code.
 */
int
unmap (ulong va, as_t *as)
{
        ulong pd_offset, pt_offset, pa, count = 0;
        pte_t *pde;
        pte_t *tva, *pgdir;
        ulong mapcnt;

        if ( as == NULL ) {
                printk ("\n PANIC: unmap: as == NULL");
                panic();
        }
        pgdir = as->pgdir;
        if ( pgdir == NULL ) {
                printk ("\n PANIC: unmap: pgdir = NULL \n");
                panic();
        }
        pd_offset = PGDIR_OFFSET(va);
        pt_offset = PGTABLE_OFFSET(va);
        pde = &(pgdir[pd_offset]);
        if ((pde->pte & PG_PRESENT) == NULL) {
                printk("\n unmap: PGDIR ENTRY NOT VALID. NON-PRESENT PAGE-TABLE \n");
                return -1;
        }
        tva = get_temp_mapping (PAGE_ALIGN(pde->pte));
        if ( tva[pt_offset].pte & PG_PRESENT == NULL ) {
                printk("\n unmap: PG_TABLE ENTRY NOT VALID. NON-PRESENT PAGE\n");
                return -1;
        }
        pa = PAGE_ALIGN(tva[pt_offset].pte);
        unregister_mapping (va, pa, as);
        return 0;
}

/*
 * This guy, handles , Kernel page faults.
 */
void kernel_pagefault_handler( ulong va )
{
        seg_t *host;
        int retval;

        if ( va < KERNEL ) { /* This is not a kernel fault */
                printk("\n Kernel PF handler called instead of user handler\n");
                panic();
        }

        host = kas.segments;
        while ( host != NULL ) {
#ifdef  __PAGEFAULTDEBUG__
        printk ("\n K_PF_H: Scanning va %x in segment base %x, limit %x \n", va, host->base , host->limit);
#endif
                if ( va_in_seg ( va , host ) ) {
                        retval = host->pageFaultHandler(va);
                        if ( retval == -1 ){
                                printk("\n PageFault at Va %x. -> Handler failed.", va);
                                panic();
                        }
                        return;
                }
                host = host->next;
        }
        printk ("\n KPFH: Unable to handle kernel page fault. panicking \n");
        printk("\n VA = %x", va);
        panic();
}

int va_in_seg ( ulong va, seg_t *host )
{
        if ( ( va >= (ulong) (host->base) )  &&  ( va < ( (ulong) (host->base) + host->limit ) ) ) return 1;
        return 0;
}

/*
 *      map_segment_handler -> This is very critical.
 *      This guy has to map the faulted va to pa.
 *      Note:
 *      1. The variable "early_stage" tells that, the kernel is still in
 *         MEMSETUP stage. hence we cant call "pmem_alloc" at this stage.
 *      2. The data structures required for pmem_alloc is being setup
 *         now. So follow a direct way of allocating memory for page tables
 *         and pa.
 */
int
map_segment_handler (ulong va)
{
        static ulong pa = NULL;
        int retval = 0;

#ifdef __PAGEFAULTDEBUG__
        printk ("\n In MAP_SEGMENT_HANDLER \n");
#endif
        va = PAGE_ALIGN(va);
        if ( early_stage ) { /* use a direct way of allocating phy mem */
                if ( !pa ) pa = LOADPOINT + KERNPAGES * PAGE;
                else pa += PAGE;
        } else {
                pa = pmem_alloc (SYSTEM);
#ifdef __PAGEFAULTDEBUG__
        printk("\n PF-Handler -> calling pmem_alloc, Not in early_stage \n");
#endif
                if ( pa == PA_MAX ) {
                        printk("\n map_seg_handler: pmem_alloc failed");
                        return -1;
                }
        }
/*
 *      Now, i got a phy memory address.
 *      Now, implement  "F(VA) = PA"
 */
#ifdef __PAGEFAULTDEBUG__
        printk ("\n Mapping Va %x to Pa %x", va ,pa);
#endif
        retval = map ( va , pa , &kas );
        return retval;
}

int panic_segment_handler ( ulong va)
{
        asm("cli");
        printk ("\n Kernel panic at VA= %x", va);
        while (1);
        return -1;
/*
 * This makes sense ? NOPE!! adding only for generality
 */
}

void
userPageFaultHandler(ulong va)
{
        printk("\n he he.. we hit the userpagefault handler :)");
        printk("\n PageFaulted address = %x", va);
        while(1);
        return;
}

int panic()
{
        asm ("cli");
        printk ("\n A kernel panic occured.. Reboot ur system \n");
        while(1);
        return -1;
/*
 * This makes sense ? NOPE!! adding for generality
 */
}
