/*                      	PMEM.C
 *                      	------
 *
 *      Author		: K.Sarnath
 *      Creation Date 	: 20-Jan-2001.
 *      	This file contains routines that manage physical memory. 
 *	This guy, manages only pages. He doesn't worry about internal 
 *	allocation of memory. As memory allocation to the kernel subsystems
 *	take place at the Virtual Address level, none wants a non-paged physical
 *	memory manager. The efficient usage of physical memory depends on the
 *	the Virtual Address allocation scheme.
 *		Continous physical memory allocation is done in a very trivial way.
 *	This algorithm can certainly be improved.
 */

#include <generic/types.h>
#include <i386/page.h>
#include <mm/pmem.h>
#include <mm/vm.h>

extern pte_t *kpgd;
extern int printk();
extern pte_t *map_page_table;
extern panic();
extern seg_t kmemseg, statickernelseg;
extern get_temp_mapping();

pmem_t *mem_table = (pmem_t *) (KERNEL + LOADPOINT + KERNPAGES*PAGE);
pmem_t *freememhead = (pmem_t *)NULL;
pmem_t *pagetable_head = (pmem_t*)NULL;
pmem_t *pagetable_tail = (pmem_t *)NULL;

ulong freemem, phymem, phypages;
int early_stage;
char *biosdata = (char *) ( KERNEL + 0x400 );
uint biosmem;
ulong check_mem (char *addr , ulong length);
ulong  estimate_memory();

/*
 *	Function Name	:	pmem_alloc
 *	Prototype	:	unsigned long pmem_alloc( int owner )
 *	Argument	:	owner -> Kernel or User.
 *	Functionality	:	This function allocates a free page.
 *				All free pages are arranged as a singly linked
 *				list. Infact the array members of the mem_table
 *				which are free are linked via a list. This function
 *				returns the head of the free list of pages.
 *	Complexity	:	O(1)
 *	Callers		:	Typically page fault handlers, xlatmgr, for allocating
 *				an unmapped page.
 */
ulong
pmem_alloc(int owner)
{
        ulong index , retval;
        pmem_t *retpage  = (pmem_t *)NULL;

        if ( freememhead == NULL ) return PA_MAX;

        retpage = freememhead;
        freememhead->status = ALLOCATED;
        freememhead->owner = owner;
        freememhead = freememhead->link.next_free;
        retpage->link.next_free = NULL;
        index = ( retpage - mem_table );
/*
 * Don't cast as ulong, compiler takes care of all the nuances involved in
 * pointer arithmetic.
 * - Sarnath.K
 */
        retval = ( index << PAGE_SHIFT );
        freemem--;
/*
 * 	Now check, whether this page is allocated for the page-table
 * If so, add it to linked list of page-table.
 * - Sarnath.K
 *	This bit of code is not at all used now. We no longer need any
 * linked list of page tables. But probably some day it might help.
 * Lets c.
 *							   Sep/12/2001.
 */
        if ( owner == PAGE_TABLE ) {
                if ( pagetable_head  == NULL ) {
                        pagetable_head = &(mem_table[index]);
                        pagetable_tail = &(mem_table[index]);
                } else {
                        pagetable_tail->link.next_page_table = &(mem_table[index]);
                        pagetable_tail = &(mem_table[index]);
                }
        }
        return retval;
}

/*
 * Function Name	:	get_nc_pages
 *
 * Prototype		:	ulong get_nc_pages( ulong n, int owner );
 *
 * Callers		:	Typically device drivers.
 *
 * Algorithm		:	Scans thru the list of pages for "n" continously free
 *				pages. Not a bit efficient.  ( 13/Sep/2001)
 * Optimization	to be
 * considered		:	Apart from maintaining free pages in a linked list, the
 *				entire physical memory can be represented as a tree of
 *				fragments.The criterion of the "tree" being the size of
 *				the free fragment of memory. That way allocation of bunch
 *				of memory would be easy.
 *
 * Functionality	:
 * 				Get N "continous" free physical page. Criterion for 
 *				continuity -> pmem_t * structures for continous pages 
 *				are also continous. pages allocated by "get_nc_pages", 
 *				shall be freed one by one, by calling pmem_free on each page.
 *
 */

ulong
get_nc_pages ( ulong n , int owner )
{
        ulong hits = 0;
        pmem_t * tr = freememhead;
        pmem_t *retval = (pmem_t*)NULL , *prev_free = (pmem_t*)NULL;

        while ( hits < n )
        {
                if ( !tr )
                        return PA_MAX;

                if ( !hits ) {
                        hits++;
                        retval = tr;
                        goto loopagain;
                } else {
                        if ( (tr - prev_free) == 1 ) {
                                hits ++;
                                goto loopagain;
                        } else {
                                hits = 0;       // Search again for continous pages.
                                retval = NULL;
                                goto loopagain;
                        }
                } //end else (!hits)
loopagain:
                prev_free = tr;
                tr = tr->link.next_free;
        } // end while ( hits < n )

        // After the "while" , retval -> Contains the address of
        // "page" structure, from where "n" continous free physical memory is
        // available.

        if ( retval == NULL )
                return PA_MAX;

        if ( freememhead == retval)  {
                ulong i;

                for ( i=0; i<n; i++) {
                        freememhead = freememhead->link.next_free;
                        if ( freememhead == NULL ) {
                                printk ("\n get_nc_pages : Freememhead is NULL after finding nc pages..Very Bad \n");
                                panic();
                        }
                }
                freemem -= n;
                return ( (retval - mem_table) << PAGE_SHIFT );
        } // end if(freememhead == retval )


        // Now , reach out to "retval" and make app changes in the
        // linked list of free physical memory.
        // The process is illustrated below.
        // freememhead -> x1 -> x2 -> x3 -> retval -> ... n times -> xz
        // I need to make this appear
        // freememhead -> x1 -> x2 -> x3 -> xz -> .....
        // Ooops, i could have had a doubly linked list, but it wouldnt be
        // memory effective. (4 bytes * NUMPHYSICALPAGES) would b 2 costly
        // for systems that have massive physical memory.

        tr = freememhead;
        while ( tr->link.next_free != retval ) {
                tr = tr->link.next_free;
                if ( tr == NULL ) {
                        printk ("\n get_nc_pages : \"tr\" is NULL after finding nc pages..Very Bad \n");
                        panic();
                }
        }
        tr->link.next_free = (retval + n - 1 )->link.next_free;
        freemem -= n;
        return ( ( retval - mem_table ) << PAGE_SHIFT );
                        //Thatz it.
}


ulong
pmem_zalloc ( int owner )
{
        ulong retval , i;
        char *tmp = (char *)NULL;

        retval = pmem_alloc ( owner );
        if ( retval != PA_MAX ) {
                tmp = (char *)get_temp_mapping ( retval );
                for (i=0;i<PAGE;i++)
                        tmp[i] = 0;
        }
        return retval;
}


int
pmem_free(ulong pa)
{
        if ( (pa & (PAGE-1)) ) {
                printk ("\nWarning: pmem_free => \"pa\" is not page-aligned");
                return -1;
        }
        pa = pa >> PAGE_SHIFT;
        mem_table[pa].status = FREE;
        mem_table[pa].owner = SYSTEM;
        mem_table[pa].link.next_free = NULL;

        if ( freememhead == NULL )
                freememhead = &(mem_table[pa]);
        else {
                mem_table[pa].link.next_free = freememhead;
                freememhead = &(mem_table[pa]);
        }
        freemem++;
        return 0;
}


/*
 * Prints PTE for a VA.
 */
void
printdebuginfo ( ulong va )
{
        ulong ptable,  pd_offset , pt_offset;
        pte_t * tva;

        pd_offset = PGDIR_OFFSET(va);
        pt_offset = PGTABLE_OFFSET(va);
        ptable = kpgd[pd_offset].pte;
        ptable = PAGE_ALIGN(ptable);
        tva = (pte_t *)get_temp_mapping ( ptable );
        tva = tva + pt_offset;
        printk ("\n PTE FOR %x  = %x",va, tva->pte);
}


/*
 * Function Name	:	memsetup
 *
 * Prototype		:	void memsetup()
 *
 * Speciality		:	1st page fault in the system happens here.
 *
 * Functionality	:	This is the function that setsup the pmem_t
 *				structure in memory. ( pmem_t structure is
 *				the data structure reqd for physical mem allocn
 *				The function deliberately page-faults at a
 *				pre-determined address. The pagefault handler,
 *				sets up mapping between the faulted address and
 *				a physical page, both increasing pagewise.
 *				The variable "early_stage" is asserted so that
 *				pagefault handler doesnt call pmem_alloc to fix
 *				a page fault rather make him allocate pmemory
 *				alternaltively. see file kvm.c for more details on
 *				the page-fault handling.
 */

void
memsetup()
{
        pmem_t *wr = (pmem_t *) NULL;
        ulong index , temp, tablesize;
        seg_t *tr = (seg_t *)NULL;
/*
 * 	This guy shall call estimate_memory, use the amt of mem,
 * 	to decide upon the size of the table.
 */
        estimate_memory();

/*
 *	Now extend the statickenelsegment and expose it as a MAP segment
 *	So that we can handle the pagefaults correctly.
 */	
        tablesize = phypages * sizeof (pmem_t);
        if ( PAGE_ALIGN(tablesize) < tablesize )
                tablesize = PAGE_ALIGN (tablesize) + PAGE;
        else
                tablesize = PAGE_ALIGN (tablesize);
        statickernelseg.limit += tablesize;
        statickernelseg.type  = MAP;
        statickernelseg.pageFaultHandler = &map_segment_handler;
	
#ifdef __DEBUG__
        printk ("\n MEMSETUP: Number of physical pages = %x",phypages);
        printk ("\n MEMSETUP: Allocating %x bytes for memory tables", phypages * sizeof (pmem_t) );
        printk ("\n Clearing the memory table to zero. Let the page fault               handler do all the tough task of mapping\n");
#endif

        early_stage = 1;
        freemem = phypages;
        pagetable_head = NULL;  // Neglect early-stage page tables.
        pagetable_tail = NULL;  // They are never de-allocated.
        wr = mem_table;
        for (index =0; index < phypages; index++) {
                wr->status = FREE;
                wr->owner = SYSTEM;
                wr->link.next_free = (wr + 1);
                wr->mapcount = 0;
                wr->mappings = NULL;
                wr++;
        }
        wr--;
        wr->link.next_free = NULL;

	statickernelseg.type = PANIC;
	statickernelseg.pageFaultHandler = &panic_segment_handler;
/*
 * Refer CONFIG.H  for details about the constants  LOWMEM, END_LOWMEM etc..
 */
        for ( index = START_LOWMEM ; index < END_LOWMEM; index += PAGE ) {
                temp = index >> PAGE_SHIFT;
                mem_table[temp].status = ALLOCATED;
                if ( index == FIRSTPTABLE || ( index >= EARLY_PTABLE && index < FREEMEM ) )
                        mem_table[temp].owner=PAGE_TABLE;
                else
                        mem_table[temp].owner =  SYSTEM;
                mem_table[temp].link.next_free = (struct page *) NULL;
                wr->mapcount = 0;
                wr->mappings = NULL;
                freemem--;
        }
        freememhead = &(mem_table[END_LOWMEM >> PAGE_SHIFT]);

/*
 * Account for the kernel stack
 */
        mem_table[(KERNEL_STACK >> PAGE_SHIFT) - 1].link.next_free = &(mem_table[( (END_KERNELPAGES+tablesize) >> PAGE_SHIFT ) + 1]);
        for ( index = KERNEL_STACK; index <= ( END_KERNELPAGES+tablesize); index += PAGE ) {
                mem_table[index >> PAGE_SHIFT].status = ALLOCATED;
                mem_table[index >> PAGE_SHIFT].owner =  SYSTEM;
                mem_table[temp].link.next_free = (struct page *) NULL;
                wr->mapcount = 0;
                wr->mappings = NULL;
                freemem--;
        }

                /* SYS stack at phy address = 0x90000 - 0xA0000 , 64k */
        mem_table[(KERNEL_STACK >> PAGE_SHIFT)].status = ALLOCATED;
        mem_table[(KERNEL_STACK >> PAGE_SHIFT)].owner = SYSTEM;

/*
 * Now update the kernel segments such that, the mem_table
 * is treated as PANIC segment. Now that mapping is set for
 * the table.
 */
        early_stage = 0;
        return;
}

/*
 *      ----------------------------------------------------
 *      This function assumes that phy mem greater than 2MB.
 *      -----------------------------------------------------
 */
ulong
estimate_memory()
{
        unsigned int retval;

        biosmem = *(uint *)(biosdata + 0x13);
        printk("\n\t\tPhysical memory manager\n");
        printk (" Estimating memory.....Please wait \n");
        printk (" This might take some time, according to memory installed. Please wait till then...\n");

        phymem = 1 << 21;       /* Assume memory is at least 2 MB */
        while (1) {
                memory_map( kpgd, 0xf0000000, phymem,1024, map_page_table);
                printk ("\n Mapped 4MB. Checking for 4MB");
                retval = check_mem( (char *)0xf0000000, 1 << 22 );
                printk ("\tCheck_mem returned %x ", retval);
                phymem += retval;
                if (retval  <  (1 << 22) ) break;
        }
        printk ("\n Phy memory installed on board (in HEX) = %x MB", phymem >> 20);
        if (phymem == 1 << 21) {
                printk ("\n System seems to have less memory..Cant run..sorry \n");
        panic();
        while (1);
        }

        phypages = phymem >> PAGE_SHIFT;
        return phymem;
}

/*
 *      This function is not a generic one.It's called only by "estimate_memory"
 *      It needs to be tailored app, b4 making it a general one.
 *      This function maps va to pa, in the given page directory.
 *      The length is in unit of pages.
 *      VA -> 4mb aligned. That is, ptable is fresh.
 *      pa -> page aligned.
 *      Here "ptable" -> Physical address.
 *
 */
memory_map ( pte_t *pgdir, ulong va, ulong pa, ulong length , pte_t *ptable)
{
        ulong pgdir_offset;
        ulong index;

        if (length > 1024) return -1;
        if (ptable == 0) {
                /* Allocate a page table. Yet to write this code */
        }
        else
        {
                pgdir_offset = PGDIR_OFFSET ( va );
                pgdir[pgdir_offset].pte = 0;
                pgdir[pgdir_offset].pte = (ulong) ptable;
                pgdir[pgdir_offset].pte |= PG_PRESENT;
                pgdir[pgdir_offset].pte |= PG_WRITE;

                ptable =(pte_t *)( (char *)ptable + KERNEL);
                for (index = 0 ; index < length; index++) {
                        ptable[index].pte = 0;
                        ptable[index].pte = pa;
                        ptable[index].pte |= PG_PRESENT;
                        ptable[index].pte |= PG_WRITE;
                        pa += PAGE;
                }
        }
        flush_tlb();
        return 0;
}

ulong
check_mem (char *addr , ulong length)
{
        uchar oldval , newval;
        ulong index = 0;

        while (index < length) {
                oldval = *addr;
                *addr  = oldval + 1;
                newval = *addr;
                if (newval == oldval) break;
                *addr  = oldval;
                addr  += PAGE;
                index += PAGE;
        }
        return index;
}
