/*
 * NOTE: set tabstop to 4 for CORRECT view.
 *
 * Copyright and Disclaimer							Sarnath 2002
 *
 *	The software is provided "AS IS" and comes with absolutely no warranty.
 *  Use it at you own risk. 
 *
 * CONTENTS
 *      All about threads ( user and kernel ).
 *                                                      --Sarnath.K
 *                                                      ( 8/May/2001 )
 *
 * A Quote from FORTUNE
 *      We learn from history that we learn nothing from history.
 *                                                      -- G.B.Shaw.
 */

#include <config.h>
#include <i386/page.h>
#include <i386/descr.h>
#include <i386/tss.h>
#include <threads/thread.h>

extern pte_t    *kpgd;
extern as_t     *cas;
extern void     userPageFaultHandler();
extern unsigned short int ldt_selector;
extern idle_thread;

seg_descr_t *gdt = (seg_descr_t*) ( KERNEL + GDTBASE );
sys_thread_t *sys_run_q = (sys_thread_t*) NULL;
sys_thread_t *current_sys_thread = (sys_thread_t*) NULL;
user_thread_t *userRunQ = (user_thread_t *)NULL;
user_thread_t *current_user_thread = ( user_thread_t *)NULL;
enum THREAD_TYPE current_type = IDLE;
unsigned int globalId = 0;

struct processInfo * getProcessInfo ( char *);

/*
 *      memcpy
 *
 *      Functionality: Copies SOURCE to DEST, LEN characters.
 *
 *      Callers:       -- NA --
 *
 *      Optimization to be considered:
 *                      Copy ULONGs by ULONGs rather than Byte by Byte.
 *
 */
int
memcpy ( char *src, char *dest, int len )
{
        int i;

        for (i=0; i<len; i++) {
                *dest = *src;
                src++;
                dest++;
        }
}

/*
 *      makeZero
 *
 *      Functionality: initializes the memory area to zero, len bytes
 *
 *      Optimization to be considered:
 *                      Take ULONG advantage.
 */
int
makeZero ( char *p, ulong len )
{
        int  i;

        for ( i=0; i<len; i++) {
                *p = 0; p++;
        }
        return 0;
}

/*
 *      create_kernel_thread:
 *
 *      Functionality:
 *              Creates a kernel thread. No LDT for such fellas.
 *      These fellas can be used to implement a seperate thread of execution
 *      in the kernel address space. I do not forsee any immediate necessity
 *      for this.
 *
 *      Caveat:
 *              Currently the CPU scheduler prioritizes a kernel thread over
 *      a user Thread. Thus a "while(1)" kernel thread would just hog the CPU
 *      for ever.
 */

void
create_kernel_thread ( ulong kva)
{
/*
 *      cli before calling me. Else Timer Interrupt may find runqueues
 *      inconsistent.
 */
        struct tss *tssp;
        char *thread_stack;
        sys_thread_t *tcb;

        ASSERT ( kva >= KERNEL )

        tcb = (sys_thread_t *) kmem_alloc(sizeof(struct sys_thread));
        if ( tcb == NULL ) {
                printk ("\n kmem_alloc:TCB==NULL:create_kernel_thread");
                panic();
        }
        tssp = (struct tss *)kmem_alloc ( sizeof (struct tss ) );
        if ( tssp == NULL ) {
                printk ("\n kmem_alloc:TSSP==NULL:create_kernel_thread");
                panic();
        }
        makeZero ( (char *)tssp , sizeof (struct tss) );
        thread_stack = (char *)kmem_alloc ( PAGE );
        if ( thread_stack == NULL ) {
                printk ("\n kmem_alloc:THREAD_STACK==NULL:create_kernel_thread");
                panic();
        }
        asm __volatile__ ("pushfl ; popl %%eax ; movl %%eax, %0"::"m" (&tssp->eflags) );
        tssp->eflags |= (1 << 9);    // set IE flag. Otherwise, we cannot 
				     // preempt this thread :)
        tssp->cr3 = KPGD;
        tssp->eip = kva;
        tssp->cs =  KERNEL_CS;
        tssp->ss0 = tssp->ss = tssp->ds = tssp->gs = tssp->fs = tssp->es = KERNEL_DS;
        tssp->ss1 = tssp->ss2 = tssp->ss0;
        tssp->esp = tssp->esp0 = (ulong)thread_stack + PAGE - 1;

        tcb->tas = &kas;
        tcb->tssp = tssp;
        tcb->next_run = tcb->prev_run = NULL;
        tcb->tq = NR_TICKS;
        add_to_st_run_queue ( tcb ); // add to system thread run queue.
}

/*
 *      add_to_st_run_queue
 *
 *      Functionality:
 *              Adds a kernel thread to the system wide runQ of kernel threads.
 *      Runnable kernel threads are arranged as a linked list.
 *
 *      Optimization:
 *                              NIL
 */
int
add_to_st_run_queue (  struct sys_thread *tcb )
{
        ASSERT ( (tcb != NULL) && ((ulong)tcb > KERNEL) )
        if ( sys_run_q == NULL ) {
                sys_run_q = tcb;
                tcb->prev_run = tcb;
                tcb->next_run = tcb;
        } else {
                tcb->next_run = sys_run_q;      // Add to head of Q.
                tcb->prev_run = sys_run_q->prev_run;
                sys_run_q->prev_run->next_run = tcb;
                sys_run_q->prev_run = tcb;
                sys_run_q = tcb;
        }
        return 0;
}

/*
 *      addToUserRunQ
 *
 *      Functionality:
 *              Adds a userThread ( a Process ) to the linked list of runnable
 *      user threads.
 *
 *      Optimization:
 *                      NIL. Its already O(1).
 */

int
addToUserRunQ ( struct user_thread *u )
{
        ASSERT( u != NULL );
        printk("\n addToUserRunQ called");
        if ( userRunQ == NULL ) {
                userRunQ = u;
                userRunQ->prev_run = u;
                userRunQ->next_run = u;
                printk("\nuserRunQ = %x", userRunQ);
        } else {
                u->next_run = userRunQ;
                u->prev_run = userRunQ->prev_run;
                userRunQ->prev_run->next_run = u;
                userRunQ->prev_run = u;
                userRunQ = u;
        }
        return 0;
}

/*
 *      get_a_user_thread
 *
 *      Functionality:
 *              Returns the head of the linked list of User Threads.
 *      Called by "schedule".
 *
 */
static inline
user_thread_t *
get_a_user_thread()
{
        return userRunQ;
}


/*
 *      get_a_sys_thread
 *
 *      Functionality:
 *              Returns the head of the linked list of Kernel Threads.
 *      Called by "schedule".
 *
 */
static inline
sys_thread_t *
get_a_sys_thread()
{
        return sys_run_q;
}

/*
 *      schedule - The MAN
 *
 *      Functionality
 *              Tries to get a runnable kernel thread. If present, schedules
 *      that thread (or) fetches a runnable user thread to schedule.
 *      The algorithm used is a simple RR.
 *
 *      Callers:
 *              From Timer Interrupt. ( kernel/386traps.c )
 *
 *      Optimization to be considered:
 *              Lot of things are there... I am against hierarchial scheduling.
 *      I feel that it doesn't model the scheduling scenario properly. There is
 *      no way u can classify processes before they start executing.. Dynamic
 *      scheduling decisions should really be considered. These issues would
 *      aggravate when sleep-wakeups come into picture. RealTime processes etc..
 *      should really be considered.
 *                                                              Sep/1/2001.
 */
int
schedule()
{
        struct sys_thread *s = NULL;
        struct user_thread *u = NULL;

schedule_again:

        switch ( current_type ) {
                case IDLE:
                        s = get_a_sys_thread ();
                        if ( s != NULL )
                                break;
                        u = get_a_user_thread();
                        printk("\nschedule: get_a_user_thread returned %x", u);
                        break;

                case SYSTEM:
                        if ( current_sys_thread->next_run )
                                move_last_sys_runqueue ( current_sys_thread );
                        current_type = IDLE;
                        goto schedule_again;

                case USER:
                        printk ("\n user ");
                        current_type = IDLE;
                        goto schedule_again;
        }

/*
 *      Here we either get a "S" or "U". If neither, just return.
 * Before dispatching, set current type appropriately.
 *                                                      -- Sarnath.K
 */

        if ( s ) {
                s->tq = NR_TICKS;
                switch_to_sys_thread(s);
                return 0;
        }

        if ( u ) {
                u->tq = NR_TICKS;
                switch_to_user_thread(u);
                return 0;
        }
        return 0;
}

/*
 *      move_last_sys_runqueue.
 *
 *      Functionality.
 */
int
move_last_sys_runqueue ( sys_thread_t *s )
{
        if ( sys_run_q ) {
                s->next_run->prev_run = s->prev_run;
/*
 * There wont be any NULL pts. See how i add to sys run q
 */
                s->prev_run->next_run = s->next_run;
                if ( sys_run_q == s )
                        sys_run_q = sys_run_q->next_run;
/*
 * Now add at the end of circular list
 */
                sys_run_q->prev_run->next_run = s;
                s->prev_run = sys_run_q->prev_run;
                s->next_run = sys_run_q;
                sys_run_q->prev_run = s;
                s->tq = NR_TICKS;
        }
}

int
moveLastUserRunQ ( user_thread_t *u )
{
        if ( userRunQ ) {
                u->next_run->prev_run = u->prev_run;
                u->prev_run->next_run = u->next_run;
                if ( userRunQ == u )
                        userRunQ = u->next_run;

                userRunQ->prev_run->next_run = u;
                u->prev_run = userRunQ->prev_run;
                u->next_run = userRunQ;
                userRunQ->prev_run = u;
                u->tq = NR_TICKS;
                return 0;
        } else {
                printk("\nmoveLastUserRunQ: userRunQ == NULL");
                return 0;
        }
}

int
switch_to_sys_thread(sys_thread_t *s)
{
        struct tss *tssp;

        ASSERT ( s != NULL )

        tssp = s->tssp;
        create_descriptor (&gdt[THREAD_TSS >> 3],(ulong)tssp,
                           sizeof(struct tss),SYSTEM_SEGMENT, KPL,
                           AVAIL_TSS_32BIT, THIRTYTWO , PAGE_GRANUALAR);
/*
 *      NOTE:  The above code is very tricky. It uses some unfamiliar and
 *      often thought to be unusable feature of 386+ machines namely
 *		descriptor caching. Actually this is an accidental discovery.
 *
 *      All tasks share the same TSS entry in the GDT. The code works cooly 
 *		because of shadow registers of 386. The TR which points to the 
 *		currently active TSS descriptor in GDT, caches the entire descriptor 
 *		inside the processor itself. This feature is taken advantage during 
 *		task switching. For example:
 *      Process A executed in the TSS context for 5 ticks, lets say.
 *      After 5 ticks, this task is pre-empted and another task ( may be
 *      the same task ) is brought in. At this time, I overwrite the
 *      TSS descriptor with TSS of new task, which apparently means that
 *      the TSS context of prev task is overWritten. But however since
 *      386+ machines cache the TSS in shadow registers, the processor
 *      saves the current state using the shadowed TSS and takes the new
 *      state from the descriptor I had created above in RAM (GDT). 
 *		Coool stuff !!
 *                                                              Sep 3/2001
 *		:ETON
 */
        current_sys_thread = s;
        current_type = SYSTEM;
        cas = &kas;
        asm ("ljmp $0x20, $0x0");
}

int
switch_to_user_thread (user_thread_t  *u)
{
        struct tss *tssp;

        ASSERT ( u != NULL )
        tssp = u->tssp;
        create_descriptor ( &gdt[THREAD_TSS >> 3], (ulong)tssp,
                            sizeof(struct tss), SYSTEM_SEGMENT,
                            KPL, AVAIL_TSS_32BIT, THIRTYTWO, PAGE_GRANUALAR);


        create_descriptor ( &gdt[ldt_selector >> 3], &(u->processLdt),
                            sizeof(ldt_t), SYSTEM_SEGMENT, KPL,
                            LDT, 0, PAGE_GRANUALAR );
/*
 *      See the comments in switch_to_sys_thread
 */
        current_user_thread = u;
        current_type = USER;
        cas = u->uas;
        printk("\n Jumping into UserSpace");
        asm ("ljmp $0x20, $0x0"); // $0x20 -> THREAD_TSS
}

int
idle_kernel_thread()
{
        while ( 1 );
}

/*
 *                      "createProcess"
 *
 *      No FORKing Fundaas. Create Task will create a new Task in the system. 
 *      THREAD mgmt can be done in user-space. Different threads of execution 
 *      sharing the same user address space, are not visible to the kernel. The
 *      kernel sees them as a single entity. Such threads can be implemented in
 *      USER level itself. In case, we need a scneario like "fork" after COW, 
 *      then just do create_process(THE SAME PROCESS). That would do. In most 
 *      UNIX system, the first thing most process do after fork is EXEC. I want
 *      to reduce the cost of FORK and EXEC with a single "create_process". 
 *
 *                                                      1/Sep/2001
 */
int
sysCreateProcess(char *processName)
{
        struct user_thread *ut;
        struct seg *textSegment , *dataSegment;
        struct tss *tssp;
        ulong pa, retval;

/*
 * processName should represent a FileSystem Object actually.
 * But Currently I m not ( cannot ) using it.
 *                                      -- Aug 19, 2001.
 */
        ut = (struct user_thread *) kmem_alloc( sizeof(struct user_thread) );
        ASSERT ( ut != NULL );
        makeZero ( (char *)ut, sizeof(struct user_thread));
#ifdef THREADDEBUG
        printk("\n An userThread was allocated");
#endif
        ut->id = globalId++;
        if ( globalId >= NR_THREADS ) {
/*
 *	Should actually change this strategy.
 */
                printk("\n globalId %x, NR_THREADS %x", globalId, NR_THREADS);
                return -1;
        }
        ut->uas = (as_t *) kmem_alloc( sizeof(as_t) );
        ASSERT( ut->uas != NULL );

#ifdef THREADDEBUG
        printk("\n An userAddressSpace was allocated");
#endif
        makeZero ( (char *)(ut->uas), sizeof(as_t) );

        ut->info = getProcessInfo( processName );
#ifdef THREADDEBUG
        printk("\n getProcessInfo returned");
#endif

        textSegment = (seg_t *) kmem_alloc( sizeof(seg_t) );
        ASSERT( textSegment != NULL );
        textSegment->base = TEXT_START;
        textSegment->limit = ut->info->textLength;
        textSegment->type = USER_TEXT;
        textSegment->pageFaultHandler = (void *)&userPageFaultHandler;
        insertSegment ( ut->uas, textSegment );

		#ifdef THREADDEBUG
        	printk("\n insertSegment returned");
		#endif
        dataSegment = (seg_t *) kmem_alloc ( sizeof(seg_t) );
        ASSERT ( dataSegment != NULL );
        dataSegment->base = TEXT_START + ut->info->textLength;
        if ( !PAGE_ALIGNED(dataSegment->base) )
                dataSegment->base = PAGE_ALIGN(dataSegment->base) + PAGE;
        dataSegment->limit = ut->info->dataLength;
        dataSegment->type = USER_DATA;
        dataSegment->pageFaultHandler = (void *)&userPageFaultHandler;
        insertSegment ( ut->uas, dataSegment );

		#ifdef THREADDEBUG
        	printk("\n insertSegment returned");
		#endif

        tssp = ut->tssp = (struct tss *) kmem_alloc(sizeof(struct tss));
        ASSERT( ut->tssp != NULL );
        makeZero( (char *)(ut->tssp), sizeof(struct tss) );
		#ifdef MALIGN_DEBUG
			printk("\n tss = 0x%x, size = 0x%x", tssp, sizeof(struct tss));
		#endif

        ut->uas->pgdir = (pte_t *) kmem_align_alloc(PAGE, PAGE);
        ASSERT( ut->uas->pgdir != NULL );
        makeZero( (char *)(ut->uas->pgdir), PAGE );
		#ifdef MALIGN_DEBUG
			printk("\n pgDir = 0x%x", ut->uas->pgdir);
		#endif

        ut->kernelStack = (char *) kmem_alloc(KERNEL_STACK_SIZE);
        ASSERT ( ut->kernelStack != NULL );
		makeZero( (char *)(ut->kernelStack), KERNEL_STACK_SIZE);
/*
 * READ THIS FIRST (if at all you want to read the entire thing)
 * ----------------------------------------------------------------------------
 * The "makeZero" is not absolutely essential at this point. "kmem_alloc" 
 * routine itself allocates and maps the required pages. This bug-fix was done 
 * when the allocator postponed the mappings to the page-fault stage. Hence 
 * these comments are REALLY OUTDATED.
 *
 *							Interesting bug fix
 *							*******************
 *  SCENARIO:
 *		Replaced the OLD bit-mapped kernel memory allocator with the newly 
 *		existing one. The "kmem_alloc" routine was thoroughly tested in 
 *		user-space. However the "kmem_align_alloc" routine was coded purely in 
 *		kernel source and was not tested in the user-space.
 *
 *	BUG FIX:
 *  	After "kmem_align_alloc" was coded, it was first tested here. (while 
 *		allocating the process page directory). As expected, the kernel 
 *		rebooted immediately while jumping into userSpace. That kicked off a 
 *		code review of "kmem_align_alloc". A bug was fixed while code reviewing 
 *		"kmem_align_alloc". However that bug will HIT only during a FREE 
 *		operation. The "reboot" bug was still haunting. After inserting a lot 
 *		of DEBUG code, it was found that the "kmem_align_alloc" code worked 
 *		fine. Later it was noted that "If "kmem_alloc" of 2 pages was made for 
 *		the process page directory and aligned externally then the code worked 
 *		fine". That got me nowhere. Interestingly when I turned off the 
 *		INTERRUPT-BIT in the TSS the "Jump" into userSpace was happening 
 *		successfully. Investigation on this line took me to the roots of the 
 *		problem. The reason was a kernel stack page-fault. Page faults in the 
 *		KERNEL STACK will give rise to recursive faults (because: The processor
 *		in the process of calling the page-fault handler will try to PUSH 
 *		registers and error code into the stack. This will stack-fault again, 
 *		again and again until the processor triple-faults and reboots.) This 
 *		error did not hit me while using the bit-mapped allocator because it 
 *		used to map PAGES for all allocations immediately. In the new 
 *		allocator, this is done in a delayed fashion. Hence the reason. The 
 *		"makeZero" above was the FIX. 
 *
 *		The skeptic reader may ask "If thats the exact reason, how did
 *		this stuff WORK with an ALLOCATION of 2 pages that was aligned 
 *		externally ?". Yeah you are right. The answer is specific to the kernel
 *		memory allocator and how it allocates memory. The allocator populates
 *		the kernel-memory (KMEM region) space to maintan state information 
 *		about the allocation. Thus such pages where the allocator populates 
 *		will get mapped automatically. When the kernel stack was allocated 
 *		using "kmem_alloc", the page containing the last 400 bytes of
 *		Kernel Stack were mapped by the allocator itself (while trying to 
 *		store meta-data about allocation). Since the kernel stack did not
 *		grow more than 400 bytes in the INTERRUPT path, the code worked fine.
 *		When I used "kmem_align_alloc" only the last 8-bytes of the kstack-tail
 *		was in the mapped page. This explains the observation.
 *	
 * Moral of the Story:
 *		a) Late Lessons about page-faulting in the kernel stack.
 *		b) Bad Coding Practice. I should have tested the "kmem_align_alloc"
 *		   in the user-space. (although it was not the originator of the bug.)
 */

		#ifdef MALIGN_DEBUG
			printk("\n KernelStack = 0x%x, size =0x%x", ut->kernelStack, KERNEL_STACK_SIZE);
		#endif

        pa = get_pa( &idle_thread, &kas );
        pa = pa - PAGE_ALIGN(pa);

        tssp->cr3 =  get_pa( ut->uas->pgdir, &kas);
        printk("\n CR3 = %x", tssp->cr3 );
        tssp->eip  = TEXT_START + pa;
        tssp->cs   = gdt_selector_alloc() | APL;
        tssp->ds   = gdt_selector_alloc() | APL;
        tssp->ss   = tssp->ds;
        tssp->gs   = tssp->ds;
        tssp->fs   = tssp->ds;
        tssp->ldt_selector = ldt_selector;
        tssp->esp  = KERNEL-4;
/*
 *      We need to have BRK value which should differentiate data and
 *      stack. This shall prevent STACK from flowing into the code and
 *      data segments.
 */


        tssp->eflags = 0;
        tssp->eflags |= (1 << 9); // set IE flag. Otherwise, we cannot preempt
                                  // this thread :)
        tssp->ss0  = KERNEL_DS;
        tssp->esp0 = (ulong)(ut->kernelStack) + KERNEL_STACK_SIZE- 10;

#ifdef THREADDEBUG
        printk("\nsysCreateProcess: TSS = %x", ut->tssp);
        printk("\nsysCreateProcess: create descriptor called");
#endif

        asm("clc");
        create_descriptor( &(gdt[tssp->cs >> 3]), 0x00, KERNEL/PAGE,
                          NONSYSTEM_SEGMENT, APL, EXEC_READ_ACCESSED, THIRTYTWO,
                          PAGE_GRANUALAR );

        asm("clc");
        create_descriptor( &(gdt[tssp->ds >> 3]), 0x00, KERNEL/PAGE,
                          NONSYSTEM_SEGMENT, APL, RDWR_ACCESSED, BIG,
                          PAGE_GRANUALAR );

        attachKernelWithUser(ut->uas);
        concatenateAddressSpaces( ut->uas, &kas );
        pa = get_pa( &idle_thread, &kas );
        pa = PAGE_ALIGN (pa);
        retval = map ( TEXT_START, pa, ut->uas);
        printk("\n EIP = %x", tssp->eip);
        addToUserRunQ( ut );
        printk("\nAdded to run queue");
        return 0;
}


/*
 * Linux Manages Segments as an AVL Tree.
 * When the number of segments grows huge, the kernel shall
 * adopt this approach. Currently segments are in a simple
 * linked list. AVL tree optimization should really be considered.
 * In the current approach, the segments are in a linked list.
 * ( not in any order )
 *                                              Aug/19/2001.
 */
int
insertSegment ( as_t *as, seg_t *seg )
{
        seg->next = as->segments;
        as->segments = seg;
        return 0;
}

struct processInfo *
getProcessInfo ( char *pName )
{
        struct processInfo *retval;

        retval = (struct processInfo*) kmem_alloc( sizeof(struct processInfo) );
        ASSERT( retval != NULL );
        retval->textLength = 0x1000;
        retval->dataLength = 0x1000;
        return retval;
/*
 *      When the kernel starts supporting ELF, this function
 *      would assume full form.
 *                                              Aug/19/2001
 */
}

/*
 *      CRITICAL - PENDING:
 *      Attaching kernel with User looks as simple as a memcpy but it has
 *      other serious implications too. Whenever a new pageTable is allocated
 *      in the kernel space ( on behalf of any process OR even without anyone's
 *		behalf), we need to update the pageDirectory of all other processes 
 *		with the address of the new pageTable. Take care.
 *      Otherwise we can do it in a delayed way.
 *                                                              1/Sep/2001.
 */
int
attachKernelWithUser ( as_t *uas )
{
        unsigned int startKernelPde , endAsPde, length;

#ifdef THREADDEBUG
        printk("\nattachWithKernel called");
#endif
        ASSERT( uas->pgdir != NULL );
        asm ("clc");
        startKernelPde = PGDIR_OFFSET(KERNEL);
        asm("clc");
        endAsPde = 0x400;       /* 1024th entry in PGDIR */
        length = endAsPde - startKernelPde;
        length = length << 2;
        asm("clc");
        ASSERT ( length <= 0x1000 );
        memcpy ( (char *)&kpgd[startKernelPde],
                 (char *)&uas->pgdir[startKernelPde],
                 length );
#ifdef THREADDEBUG
        printk("\nattachWithKernel successful");
#endif
        return 0;
}

int
concatenateAddressSpaces(as_t *headAs, as_t *tailAs)
{
        seg_t *seg = headAs->segments;

        if ( seg == NULL ) {
                headAs->segments = tailAs->segments;
                return 0;
        }
        while ( seg->next != NULL )
                seg = seg->next;
        seg->next = tailAs->segments;
        return 0;
}
