/*************************************************************************
 *
 * 	NOTE: SET TAB SPACE TO 4 FOR PROPER ALIGNMENT
 *
 *	COPYRIGHT AND DISCLAIMER 								Sarnath.K	2002
 *	The software is provided "AS IS" without any warranty. You are free to 
 *	distribute and re-use the code for any purpose except commercial purpose 
 *	without hiding the author(s)'s identity and all COPYRIGHT information. 
 *	However I strongly disclaim any explicit or implicit usability of the code.
 *	I wont be responsible to any damage caused due to the study or use of this 
 *	code. Use it at your own risk. Thanks for your patience.
 *
 *	CONTENTS, COMMENTS, INFORMATION and NOTEs
 *	This file contains routines that register and unregister mappings. The 
 *	functions "register_mapping" and "unregister_mapping" are the interface 
 *	functions given to the kernel. "unmapped pages" are the hosts of the PTOV 
 *	and VTOP information and the structures "ptov_t" and "vtop_t" should be 
 *	equal in size. See file include/mm/ptov.h and include/mm/vtop.h.
 *
 *	Each time I call "get_temp_mapping", TLB is flushed and subsequently
 * 	slows down things. That really pains.  There is an instrn called "invlpg" 
 *	which is used to selectively flush the TLB. However this instruction is
 *	is available only for 486/+ processors.
 *
 *	IMPORTANT NOTE -- Sarnath.K 30/Apr/2001.
 *	Never use "get_pa" to retrieve the physical address of a temporary
 *	mapping. Temporary mapping is any mapping got by using "get_temp_mapping"
 *	This is coz "get_pa" itself uses "get_temp_mapping" to read page-table.
 *	Hence "get_pa" would corrupt any temporary mapping. So this function wont
 *	give any correct value for temporary mappings.
 *
 * REVISION HISTORY
 *
 * Sarnath		April 2001		Created, Coded, Tested the basic functionality.
 *								The layer allows to register mappings at
 *								the BYTE level. This is majorly done in
 *								order to offload the burden of Page-Level
 *								tracking of VA from the bitmapped KMEM 
 *								allocator.
 *
 * Sarnath		11-August-2002	Converted this layer to operate only at the
 *								Page Level. The effort was triggered by the
 *								new kernel memory allocator in place. The
 *								BIT mapped allocator is no more. The new
 *								allocator registers mappings only at the PAGE
 *								LEVEL. Thus we avoid the traversal of linked
 *								list of physical addresses everytime a mapping
 *								is registered.
 *								
 *****************************************************************************/

#include <mm/pmem.h>
#include <mm/ptov.h>
#include <mm/vtop.h>

static inline int unload_pagetable ( ulong pde, as_t *owner );
ptov_t * get_a_ptov();

extern int early_stage;
unmapped_page_t  *start_unmapped_page = (unmapped_page_t *)NULL;

/*************************************************************************
 *
 *  VARIABLE DESCRIPTION
 *
 *	VARIABLE	:	"start_unmapped_page"
 *
 *	"start_unmapped_page" -> Has the physical address
 * 	of the start of the unmapped pages. All such unmapped pages are in
 * 	the linked list. see defn of "unmapped_page_t" in include/mm/ptov.h
 *
 **************************************************************************/

ulong num_unmapped_pages;

#ifdef __XLATDEBUG__
        ulong get_ptov, free_ptov;
#endif


/***********************************************************************
 *
 *	FUNCTION		:	register_mapping
 *
 *	TYPE			:	Interface function to the rest of the kernel.
 *
 *	FUNCTIONALITY 	:
 * 	This function registers the "translation", that has exists in an address 
 *	space. The translation has already been made. Now we just need to register 
 *	it. The "pmem_t" structure that describes a page has a linked list of 
 *	structures that describe the mapping of this page in an address space. Each
 * 	structure registers a mapping (as,va) to that pa. So register the current 
 *	mapping by adding one more node in the linked list.
 *
 * REVISION HISTORY :
 *						Sarnath	2002	Created and coded the basic
 *										functionality.
 *	
 *********************************************************************/

int
register_mapping(ulong va, ulong pa, struct as *owner)
{
        pmem_t *pp;
        pte_t *pgdir;
        ptov_t *mapinfo = NULL, *tmp, *tva = (ptov_t *)NULL;
        ulong tpa;


        if ( !early_stage ) {
/*-------------------------------------------------
 * Now, store the information about the physical page
 * that is being mapped.
 *-----------------------------------------------*/
                ASSERT ( sizeof(ptov_t) == sizeof(vtop_t) )
				if ( va & (PAGE-1) )
				{
					printk("\nRegister Mapping called with a non-page address");
					panic();
				}
                pp = &mem_table[ pa >> PAGE_SHIFT ];
                if ( pp->mappings == NULL ) {
                        mapinfo = get_a_ptov(owner, va);
                        if ( (ulong)mapinfo == 0xffffffff ) {
                                printk ("\n Ooops, get_a_ptov failed. Unable to register mapping ");
                                return -1;
                        }
                        pp->mappings = (void *)mapinfo;
                } else {
                        mapinfo = get_a_ptov(owner, va);
                        if ( (ulong)mapinfo == 0xffffffff ) {
                                printk ("\n Ooops, get_a_ptov failed. Unable ");
								printk (" to register mapping.");
                               	panic();
                        }
                        tva = (ptov_t *)get_temp_mapping ((ulong)mapinfo);
						tva->next = (ptov_t *)(pp->mappings);
                        tva =
							  (ptov_t*)get_temp_mapping((ulong)(pp->mappings));
                        tva->prev = mapinfo;
						pp->mappings = (void *)mapinfo;
                }
                pp->mapcount ++;
/*-------------------------------------------------------------------
 *      Till now we have registered PTOV information. Now register the
 *      VTOP information.
 *                                                      -- Sarnath.K
 *-----------------------------------------------------------------*/
                register_vtop (va, owner);
        } // END of if ( !early_stage )
        return 0;
}

/***************************************************************************
 *	FUNCTION: register_vtop
 *
 *	FUNCTIONALITY:
 *	This function registers the MAPPING in the page table's "pmem_t" structure.
 *	This helps in freeing the associated PAGE-TABLE when it services zero
 *	mappings in an address-space. Read the following scenario for an interesting
 *	discussion.
 *
 *	ASSUMPTION:
 *	Assumes that the page table is in RAM. Callers should make sure that the
 *	mappings are in place before calling this function.
 *
 *	INTERESTING SCENARIO:
 *	What if a physical page functions as a PAGE_TABLE in one address space and
 *	is mapped by a VA in some other address space. Although this scenario seems
 *  rare, it can be encountered in an UNIX box. (Thanks to Deepak Veliath)
 *	It can happen in UNIX boxes where the Super-User can open("/dev/pmem") and
 *	MAP physical pages under his address-space. The physical pages can include 
 *	PAGE-TABLES of some other processes also.
 *	Let us consider the following sequence of steps:
 *	a) A Physical Page is allocated as a PAGE-TABLE. (to serve a bunch of 
 *	   mappings in an address space.) The "register_vtop" function will register
 *	   the various mappings done via this page-table for that address-space in
 *	   the "mappings" field of the "pmem_t" structure corresponding to that
 *	   page table. Thus the "mappings" field will be seen as a "vtop_t" pointer
 *	   by this function.
 *	b) The same physical page is now mapped into another address space. The 
 *	   "register_mapping" function treats the "mappings" field as "ptov_t" 
 *	   structure and adds the current mapping to the head of the linked list. 
 *	   Since PTOV and VTOP structures host the "next" and "prev" pointers at 
 *	   the same offset, the linked list addition does not pose any problem.
 *	c) Now consider the UNMAPPING scenario. Let us assume that the address
 *	   space using the PAGE-TABLE is destroyed. As the UNMAPPING happens, the
 *	   "unregister_mapping" will be called for all the pages. The function will
 *	   decrement the reference count (The number of mappings that is serviced
 *	   by this page-table for that address-space). When it drops to zero, the 
 *	   entry will be removed from the linked-list. However the PAGE still 
 *	   retains the information that it is still mapped as a PAGE under some 
 *	   other address-space. Thus the PAGE will not be freed. When the other 
 *	   address-space relinquishes this PAGE, it will be returned to the 
 *	   free-pool of pages. By this time, A skeptic reader will certainly raise 
 *	   doubts about the co-existence of PTOV and VTOP information in a single 
 *	   linked list.
 *	   The explanation follows:
 *	i) There exists no confusion because of PTOV and VTOP structures coexisting 
 *	   in a single linked list. This is because both of these structures are 
 *	   similar in size and host the "next" and "prev" pointers at the same 
 *	   offset. They differ only by their first member. For PTOV - It is the VA,
 *	   For V-TO-P, it is the PDE. The VA is always PAGE-ALIGNED and PDE ranges 
 *	   only from (0-1023). Thus there is no possibilty of an <as,pde> being 
 *	   same as <as,va>. These reasons mandate the symmetricity of PTOV and VTOP
 *	   structures.This has been well-stressed in their respective data 
 *	   structural definitions.
 * ii) Moreover the function "get_a_ptov" is overloaded to do the function of 
 *     "get_a_vtop" also. The symmetricity of the structures has been leveraged
 *     for this.
 *
 **************************************************************************/

int
register_vtop ( ulong va, as_t *owner )
{
        ulong pde, ptable;
        pmem_t *pp;
        vtop_t *new, *tva, *tmp;

        ASSERT ( owner != NULL )
        pde = PGDIR_OFFSET(va);
        ptable = owner->pgdir[pde].pte;
        ptable = PAGE_ALIGN(ptable);

        pp = &(mem_table [ ptable >> PAGE_SHIFT ]);
        if ( pp->mappings == NULL ) {
                new = (vtop_t*) get_a_ptov ( owner,pde );
                tva = (vtop_t*)get_temp_mapping (new);
                tva->refcnt = 1;
                tva->next = tva->prev = NULL;
                pp->mappings = (void *)new;
                return 0;
        }
        else
        {
                tmp = (vtop_t*) pp->mappings;
                while ( tmp ) {
                        tva = (vtop_t*)get_temp_mapping( tmp );
                        if ( (tva->owner == owner) && (tva->pde == pde) ){
                                tva->refcnt++;
                                return 0;
                        }
                        if ( tva->next == NULL )
                                break;
                        tmp = tva->next;
                }
/*
 *	Here TMP contains Physical address and TVA contains the Virtual address of 
 *	the last node of the linked list. TVA -> Temporary mapping. "get_pa" 
 *	corrupts TVAs mapping. So "get_pa(tva,kas)" never returns a correct 
 *	physical address, coz "tva" is a temporary mapping. See the comments at the
 *	begining of the file.
 *
 */
                new = (vtop_t*)get_a_ptov ( owner, pde);
                tva = (vtop_t*)get_temp_mapping (new);
                tva->refcnt = 1;
                tva->prev = tmp;
                tva->next = NULL;
                tva = (vtop_t*)get_temp_mapping (tmp);
                tva->next = new;
        }
}

/*****************************************************************************
 *	FUNCTION:	unregister_mapping
 *
 *	FUNCTIONALITY:
 *	This function removes the NODE describing the "mapping", from the "mappings"
 *	linked list present in the "pmem_t" structure of the PAGE. It further calls
 *	"unregister_vtop" to update the database of the PAGE-TABLE serving the
 *	mapping.
 *
 *	INCONSISTENCY IN API: PLEASE NOTE:
 *	There is an inconsistency in the API exported by this function and the
 *	"register_mapping" function. This function UNLOADS the mapping automatically
 *	The callers need not do that. However the behaviour of "register_mapping" 
 *	is the other way round. The "register_mapping" expects the caller to have 
 *	already established the mapping before calling it.
 *	
 *****************************************************************************/
ulong
unregister_mapping ( ulong va, ulong pa, as_t *owner )
{
        pmem_t *pp;
        ptov_t *tmp, *tva, *prev, *next;

		if (va & (PAGE-1))
		{
				printk("\nunregister_mapping: \"va\" is not PAGE-ALIGNED");
				panic();
		}

        pp = &(mem_table[ pa >> PAGE_SHIFT ]);
        tmp = (ptov_t*) pp->mappings;
        while ( tmp ) {
                tva = (ptov_t*) get_temp_mapping ( (ulong)tmp );
                if ( (tva->owner == owner) && (tva->va == va) ) {
                        pp->mapcount --;
                        if ( pp->mapcount < 0 )
						{
                          printk ("\nWARNING:unregister_mapping:mapcount < 0");
						}
                        prev = tva->prev;
                        next = tva->next;
                        if ( prev != NULL ) {
                                tva = (ptov_t*) get_temp_mapping((ulong)prev);
                                tva->next = next;
                        } else {
                                pp->mappings = (void*) next;
                        }

                        if ( next != NULL ) {
                                tva = (ptov_t*) get_temp_mapping((ulong)next);
                                tva->prev = prev;
                        }
                        free_a_ptov (tmp);
                        unload_mappings ( va, pa, owner );
/*
 *      Till now we unregistered a PTOV information. Now since we have
 *      unloaded a mapping. So inform the VTOP layer to unregister the VTOP if 
 *      necessary.
 */
                        unregister_vtop( va, owner );
                        if (pp->mappings == NULL )
						{
                                pmem_free(PAGE_ALIGN(pa));
						}
                        return pp->mapcount;
                }
                tmp = tva->next;
        }
        printk ("\n PANIC: unregister_mapping: Unable  to find a mapping");
	   	printk (" address space. as=%x, va=%x, pp->owner = %x", owner, va,
																pp->owner);
#ifdef __XLATDEBUG__
        printk("\n get_ptov = %x, free_ptov = %x", get_ptov, free_ptov);
#endif
        panic();
}

/*****************************************************************************
 * FUNCTION: unregister_vtop
 *
 * FUNCTIONALITY:
 * This function checks whether the PAGE TABLE used in the "mapping" serves
 * any more mapping. If it does not support any more mappings then the node
 * correspoding to that <address-space, PDE> is removed from the linked list
 * of mappings. If the linked-list is empty then the function proceeds to
 * return the page to the free-physical-memory pool.
 *
 * ASSUMPTION:
 * This function does not check whether the corresponding PAGE-TABLE resides
 * in RAM OR not. It just assumes that they are all in place. The higher
 * level callers need to make sure that the PAGE-TABLES are in PHYSICAL MEMORY
 * before calling this function.
 *****************************************************************************/
int
unregister_vtop ( ulong va, as_t *owner )
{
        ulong pde, ptable;
        pmem_t *pp;
        vtop_t *new, *tva, *tmp, *next, *prev;

        ASSERT ( owner != NULL )
        pde = PGDIR_OFFSET(va);
        ptable = owner->pgdir[pde].pte;
        pp = &( mem_table[ptable >> PAGE_SHIFT] );
        tmp = (vtop_t *) pp->mappings;
        while ( tmp ) {
                tva = (vtop_t*) get_temp_mapping ( tmp );
                if ( (tva->owner == owner) && (tva->pde == pde) ) {
                        tva->refcnt--;
                        if ( tva->refcnt )
						{
                        	return 0;
						}
/*
 *      Now its time to unload the mapping for page table in (as,pde)
 *      But still, other address-spaces may still own it. Hence we cannot
 *      free the page. Although sharing of page-tables between address-spaces
 *      might seem wierd, it is still possible. This layer will support the
 *     	existence of all sort of mappings that are possible under VM semantics.
 */
                        unload_pagetable ( pde, owner );
/*
 *      Now, free the PTOV structure and update linked list.
 *      If there are no more mapping address spaces for the page-table,
 *      then we can pmem_free this Page-Table.
 */
                        next = tva->next;
                        prev = tva->prev;
                        if ( prev == NULL )
                                pp->mappings = (void *)next;
                        else {
                                tva = (void *)get_temp_mapping (prev);
                                tva->next = next;
                        }
                        if ( next != NULL ) {
                                tva = (void *)get_temp_mapping (next);
                                tva->prev = prev;
                        }
                        free_a_ptov (tmp);
                        if ( pp->mappings == NULL ) {
                                /* Now we can free the Page-Table itself */
                                ptable = PAGE_ALIGN(ptable);
                                pmem_free ( ptable );
                        }
                        return 0;
                }
                tmp = tva->next;
        }
        printk ("\nunregister_vtop: Unable to find \"as\" for this page-table");
        printk ("\n Some critical error has happened");
        ASSERT ( 1 == 0 )
        return 0;
}

static inline int
unload_pagetable ( ulong pde, as_t *owner )
{
        ASSERT (owner != NULL)
        owner->pgdir[pde].pte = 0;
        return 0;
}

int
unload_mappings ( ulong va, ulong pa, as_t *owner )
{
        ulong pd_offset, pt_offset;
        pte_t *pgdir, *ptable, *tva;

        if ( owner == NULL ) {
                printk ("\n unload_mappings: address space is NULL");
                panic();
        }
        pd_offset = PGDIR_OFFSET ( va );
        pt_offset = PGTABLE_OFFSET ( va );
        pgdir = owner->pgdir;
        ptable = (pte_t *)(pgdir[pd_offset].pte);
        tva = (pte_t*) get_temp_mapping ( PAGE_ALIGN((ulong)ptable) );
        tva[pt_offset].pte = 0;
        // We had turned off the mapping in the page-table.
        return 0;
}

ptov_t *
get_a_ptov(as_t *owner, ulong va )
{
        uchar byte , bit ;
        unmapped_page_t *tva, *tmp, *tpa; // tmp holds PA, tva holds its VA
        ptov_t *slot;
        int index;
        ulong retval, new_page;

#ifdef  __XLATDEBUG__
        get_ptov++;
#endif
        if ( start_unmapped_page ) {
                tmp = tpa = start_unmapped_page;
                while ( tmp ) {
                        tva = (unmapped_page_t*)get_temp_mapping ( (ulong)tmp );
                        if ( tva->entries == NUM_PTOVS )
                                goto search_next;
                        index = 0;
                        retval = (ulong) (tmp + 1);
                        while ( index <= (NUM_PTOVS/8) ) {
                                if ( tva->bitfield[index] == 0xff ) {
                                        index ++;
                                        retval += (8 * sizeof( struct ptov ));
                                        continue;
                                }
/*
 *      Consider making this bit-field as "ulong".
 *      Reap the real 32-bit benefit.
 *      I'll do it someday.
 *              -Sarnath.K
 */
                                break;
                        }
                        byte = tva->bitfield[index];
                        bit = 1;
                        while ( bit ) {
                                if ( (byte & bit) == 0 ) {
                                        tva->bitfield[index] |= bit;
                                        tva->entries++;
                                        slot = (ptov_t*)get_temp_mapping (retval);
                                        slot->owner = owner;
                                        slot->va = va;
                                        slot->next = NULL;
                                        slot->prev = NULL;
                                        slot->refcnt = 1;
                                        return ((ptov_t*)retval);
                                }
                                bit <<= 1;
                                retval += sizeof (struct ptov);
                        }
                        printk ("\nError in getptov.This shouldnt b executed");
                        return (ptov_t *) 0xffffffff;
                search_next:
                        tpa = tmp;
                        tmp = tva->next;
                }

/*      
 *      We reach here if all unmapped pages are packed fully.
 *      So, allocate a page, and set up the linked list.
 *      "pmem_zalloc" (and even get_pa ) contaminates "tva"'s mapping.
 *      So be careful to save and restore temporary mappings , in case we
 *      need them.
 *                                              -- Sarnath.K
 */
                 tmp = tpa;
/* Save Physical address. No point in saving
 * virtual address of a temporary mapping.
 */
                 new_page = pmem_zalloc ( SYSTEM );
                 if ( new_page == 0xffffffff ) {
                         printk ("\n pmem_zalloc failed while allocating new unmapped page ");
                         return ((ptov_t*)0xffffffff);
                 }
                 num_unmapped_pages++;
#ifdef __XLATDEBUG__
                printk ("\n An unmapped page was allocated %x", new_page);
                printk ("\n Num unmapped pages in system = %x", num_unmapped_pages);
#endif
                 tva = (unmapped_page_t*) get_temp_mapping ( (ulong)tmp );
                 if ( tva->next != NULL ) {
                        printk ("\n Some stale value in next field in the last
                                    unmapped page !");
                        panic();
                 }
                 tva->next = ( unmapped_page_t *) new_page;
                 tva = (unmapped_page_t*) get_temp_mapping ( new_page );
                 tva->entries=1;
                 tva->bitfield[0] = 1;
                 tva++;
                 retval = (ulong) ((unmapped_page_t*)new_page + 1);
                 slot = (ptov_t*)tva;
                 slot->owner = owner;
                 slot->va = va;
                 slot->next = NULL;
                 slot->prev = NULL;
                 return ( (ptov_t*)(retval) );
        } else {
                start_unmapped_page = (unmapped_page_t *)pmem_zalloc( SYSTEM );
                if ( (ulong)start_unmapped_page == 0xffffffff ) {
                        printk("\n get_a_ptov: Page allocation failed");
                        panic();
                }
                tva = (unmapped_page_t*)get_temp_mapping((ulong)start_unmapped_page);
                num_unmapped_pages++;
#ifdef __XLATDEBUG__
                printk ("\n A start_unmapped page was allocated %x\n", start_unmapped_page);
#endif
                tva->entries = 1;
                tva->bitfield[0] = 1;
                retval = (ulong) (start_unmapped_page + 1);
                tva++;
                slot = (ptov_t*)tva;
                slot->owner = owner;
                slot->va = va;
                slot->next = NULL;
                slot->prev = NULL;
                return ( (ptov_t*)(retval) );
        }
        return NULL;
}

int
free_a_ptov ( ptov_t *slot )
{
        unmapped_page_t *host, *vhost, *next, *tmp, *tva;
        ptov_t *first_slot;
        unsigned char nthbyte, nthbit, thebit;
        ulong n;


#ifdef  __XLATDEBUG__
        free_ptov++;
#endif
        if ( slot == NULL ) {
                printk("\n free_ptov: slot = NULL. Panicking !! ");
                panic();
        }

        host = (unmapped_page_t*) PAGE_ALIGN ( (ulong)slot );
        first_slot = (ptov_t*) ( host + 1);
        n = slot - first_slot;
        nthbyte = n / 8;
        nthbit = n % 8;
        thebit = ( 1 << nthbit );
        thebit = ~thebit;
        vhost = (unmapped_page_t*)get_temp_mapping ( (ulong)host );
        vhost->bitfield[nthbyte] &= thebit;     // Mark it free
        vhost->entries --;
        if ( vhost->entries == 0 ) {
                next = vhost->next;
                if ( host == start_unmapped_page ) {
                        start_unmapped_page = next;
                        pmem_free((ulong) host);
                        if ( num_unmapped_pages == 0 )
						{
                                printk ("\n free_a_ptov: Num_unmapped_pages already 0 :-(");
						}
                        else
                                num_unmapped_pages--;
#ifdef __XLATDEBUG__
                printk ("\n A start_unmapped_page %x was freed \n", host);
#endif
                        return 0;
                }
                tmp = start_unmapped_page;
                while ( tmp ) {
                        tva = (unmapped_page_t*) get_temp_mapping(tmp);
                        if ( tva->next == host )
                                break;
                        tmp = tva->next;
                }
                tva->next = next;
                pmem_free((ulong)host);
                if ( num_unmapped_pages == 0 )
                        printk ("\n free_a_ptov: Paranoid. Num_unmppaed_pages already 0 :-(");
                else
                        num_unmapped_pages--;

#ifdef __XLATDEBUG__
                printk ("\n An unmapped page was freed %x\n", host);
                printk ("\n Page b4 this page = %x", tmp);
#endif
        }
        return 0;
}

#ifdef __XLATDEBUG__

void print_seg_mappings (as_t *as, seg_t* seg );
void print_mappings ( ulong pa );

void
print_as_mappings(as_t *as)
{
        seg_t *seg;

        printk("\n PRINT_AS_MAPPINGS called ");
        if ( as == NULL ) {
                printk("PANIC: print_as_mappings : as = NULL");
                panic();
        }
        seg = as->segments;
        while ( seg ) {
                print_seg_mappings(as, seg);
                seg = seg->next;
        }
}

void
print_seg_mappings (as_t *as, seg_t* seg )
{
        ulong startva, endva, pa;

        startva = (ulong) seg->base;
        endva = startva + seg->limit;

        while ( startva < endva ) {
                pa = get_pa ( startva, as);
                if ( pa != 0xffffffff )
                        print_mappings ( pa );
                startva += PAGE;
        }
}

void
print_mappings ( ulong pa )
{
        pmem_t *pp;
        ptov_t *tmp, *tva;

        pa = pa >> PAGE_SHIFT;
        pp = &(mem_table[pa]);
        if ( (tmp = pp->mappings) ) {
                printk("\n Pa= %x, Mapping Domains=> ", pa << PAGE_SHIFT);
                while ( tmp ) {
                        tva = (ptov_t*) get_temp_mapping ((ulong) tmp);
                        printk("{ as=%x , va=%x , next=%x, prev = %x, host=%x} ", tva->owner, tva->va, tva->next, tva->prev, tmp);
                        tmp = tva->next;
                }
        }
        return;
}
#endif
